package org.ow2.choreos.iots.datatypes;



import org.json.simple.JSONObject;
import org.ow2.choreos.iots.common.Constants;
import org.ow2.choreos.iots.semantics.SemanticClass;


import java.io.Serializable;
import java.util.Date;

/**
 * This class carries should be extended by developers of new sensor drivers, to
 * store any data (measured by the sensor) that is associated with the same
 * timestamp.
 */
public abstract class SensorData extends SemanticClass implements Serializable {

	private static final long serialVersionUID = 1963099439953813653L;
	private static final String T = Constants.TVAL;
	protected SensorData() {
		
	}
	
	protected SensorData( JSONObject j) {
			super(j);
		 this.timestamp = Long.parseLong(j.get(T).toString()); 
	
		
	}
	
	/**
	 * The time in nanoseconds at which the event happened. 
	 */
	public long timestamp;

	/**
	 * Creates a new data sample that carries no data (since this is an abstract
	 * class), only a timestamp. When subclassing this, be sure to call this
	 * constructor from within any constructor that initializes data.
	 */
	protected SensorData(long timestamp) {
		this.timestamp = timestamp;
	}

	/**
	 * Returns whether the data was sampled sometime "within the near past" or
	 * if it was reused from the previous available value (this is good for slow
	 * sensors, such as GPS, where the previous value may still be useful).
	 */
	public abstract boolean isFresh();

	/**
	 * Writes this data in a human-readable format.
	 */
	public abstract String toString();

	/**
	 * Copies all the data contained in this object into a new instance, and
	 * returns it. If the class is immutable, the current instance
	 * can just be passed directly without any side-effects.
	 */
	public abstract SensorData copy();

	/**
	 * Returns an array of strings containing the name of each component of this
	 * SensorData. 
	 *
	 * No spaces allowed.
	 *
	 * For instance:
	 *
	 *   String s = new String[3];
	 *   s[0] = "position-x";
	 *   s[1] = "position-y";
	 *   s[2] = "timestamp";
	 *   return s;
	 */
	public abstract String[] getComponentNames();

	/**
	 * Writes the components of this SensorData into an array of CSV-compatible 
	 * strings (i.e. they don't contain spaces, commas, semicolons, or quotations).
	 *
	 * For instance:
	 *
	 *   String s = new String[3];
	 *   s[0] = Double.toString(x);
	 *   s[1] = Double.toString(y);
	 *   s[2] = Long.toString(timestamp);
	 *   return s;
	 */
	public abstract String[] toStringArray();

	/**
	 * Creates a data carrier from an array of CSV-compatible strings.
	 */
	public abstract SensorData fromStringArray(String[] sa);
	
	
	@Override
	public JSONObject serializeAsJSON() {
		JSONObject ret = new JSONObject();

		ret.put(T, this.timestamp);

		return ret;
	}


	
}

