/******************************************************************
Author:  		Georgios Mathioudakis (georgios.mathioudakis@inria.fr)
Project: 		PhoneServicesProxy
package: 		fr.inria.arles.websockets
Created: 		12 fevr. 2013
Updated: 		
Status:  		Working Draft
Description: 	This class implements the WebSocket server used for the streaming feature. More information can be found in https://github.com/TooTallNate/Java-WebSocket
 ******************************************************************/

package fr.inria.arles.websockets;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import net.vtrip.things.Static_Addresses;
import org.java_websocket.WebSocket;
import org.java_websocket.handshake.ClientHandshake;
import org.java_websocket.server.WebSocketServer;
import org.codehaus.jettison.json.JSONArray;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;

public class WebSocket_Server extends WebSocketServer {

	private int status = 0; // Status code 0 defines a stopped server - Status
							// code 1 defines a running server
	/*
	 * Implement the singleton design pattern. This way any class can have to a
	 * single instance of the WebSocket server class without the need of
	 * initializing it.
	 */
	private static WebSocket_Server instance = null;

	public static WebSocket_Server getInstance() {
		if (instance == null) {
			try {
				instance = new WebSocket_Server(Static_Addresses.WS_SERVER_PORT);
			} catch (UnknownHostException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return instance;
	}
	
	private Map<String, WebSocket> _registeredDevices = null; 		//Store here the registered sensor devices to be able to communicate with them later
	private Map<String, Map<WebSocket, String>> _requests = null;	//Store the requested services here as per device Map<DeviceID, Map<Websocket, service>>

	public WebSocket_Server(int port) throws UnknownHostException {
		super(new InetSocketAddress(port));
		_registeredDevices 	= new HashMap<String, WebSocket>();
		_requests 			= new HashMap<String, Map<WebSocket, String>>();
	}

	public WebSocket_Server(InetSocketAddress address) {
		super(address);
	}

	/**
	 * Starts the server after checking if it is already started previously
	 */
	public void WebSocket_Server_Start() {
		if (status == 0) {	
			status = 1;
			try{
			this.start();
			} catch(Exception e) {
				System.out.println("Error on server start: " + e.getMessage());
			}
			System.out.println("Server started on: "
					+ this.getAddress().getHostName());
		}
	}

	/**
	 * Stops the server after checking if it is already stopped previously
	 */
	public void WebSocket_Server_Stop() {
		if (status == 1) {
			try {
				status = 0;
				this.stop();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				System.out.println(e.getMessage());
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				System.out.println(e.getMessage());
			}

			System.out.println("Server Stopped");
		}
	}

	/**
	 * @return Status code 0 --> Server stopped. Status code 1 --> Server is
	 *         running
	 */
	public int Get_Server_Status() {
		return status;
	}

	@Override
	public void onOpen(WebSocket conn, ClientHandshake handshake) {
		System.out.println("WS client is connected");
	}

	@Override
	public void onClose(WebSocket conn, int code, String reason, boolean remote) {
		for(Entry<String, Map<WebSocket, String>> pair : _requests.entrySet()){
			if(pair.getValue().containsKey(conn)){
				pair.getValue().remove(conn);
				
				if(pair.getValue().size() == 0){
					_registeredDevices.get(pair.getKey()).send("{\"proxyStopSensing\":\"proxyStopSensing\"}");
					System.out.println("Send stop sensing event to device: " + pair.getKey());
				}
			}
		}
	}

	@Override
	public void onMessage(WebSocket conn, String message) {
		String deviceID = "";
		String service 	= "";

		JSONObject msgObj;
		try {
			msgObj = new JSONObject(message);
			
			/*
			 * Firstly check the type of message received. There are 3 types of messages:
			 * 1. Request for data
			 * 2. Respond with data
			 * 3. Registration of a new sensor device
			 */
			if(msgObj.has("sensorDataRequest")){ //Request from a client
				JSONObject value 	= (JSONObject) msgObj.get("sensorDataRequest");
				deviceID 			= value.getString("deviceID");
				service 			= value.getString("serviceName");
				//rate 				= value.getString("rate");
				
				System.out.println("Request for device: " + deviceID + " - service: " + service);
				
				if(_requests.containsKey(deviceID)){
					System.out.println("_requests contain device: " + deviceID);
					_requests.get(deviceID).put(conn, service);
				} else {
					System.out.println("This is a request for a new device: " + deviceID);
					HashMap<WebSocket, String> request = new HashMap<WebSocket, String>();
					request.put(conn, service);
					_requests.put(deviceID, request);
				}
				
				/*
				 * Forward the sensor data request to the specified device
				 */
				if(_registeredDevices.containsKey(deviceID))
					_registeredDevices.get(deviceID).send(message);
				
			} else if(msgObj.has("sensorDataResponse")){ //Data from sensor device
				JSONObject value 	= (JSONObject) msgObj.get("sensorDataResponse");
				deviceID 			= value.getString("deviceID");
				service 			= value.getString("serviceName");
				
				System.out.println("New data response from device: " + deviceID + " - service: " + service);
				if(_requests.containsKey(deviceID)){
					
					Map<WebSocket, String> requestConnections = _requests.get(deviceID);
					for(Map.Entry<WebSocket, String> pair : requestConnections.entrySet()){
						
						if(pair.getValue().equals(service)){
							pair.getKey().send(message);
							System.out.println("Message has been send to client");
						}
					}
				}
			} else if(msgObj.has("registerDevice")){ //New device is being registered
				JSONObject value 	= (JSONObject) msgObj.get("registerDevice");
				deviceID 			= value.getString("deviceID");
				_registeredDevices.put(deviceID, conn);
				System.out.println("Device registration: " + deviceID);
			} else if(msgObj.has("unRegisterDevice")){ //Device is being unregistered
				System.out.println("unRegisterDevice");
				JSONObject value 	= (JSONObject) msgObj.get("unRegisterDevice");
				deviceID 			= value.getString("deviceID");
				if(_registeredDevices.containsKey(deviceID)){
					_registeredDevices.remove(deviceID);
					System.out.println("Device removed: " + deviceID);
				}
			} else { return; }
			
		} catch (JSONException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Override
	public void onError(WebSocket conn, Exception ex) {

	}

	/**
	 * Sends <var>text</var> to all currently connected WebSocket clients to
	 * this specific interface for example location.
	 * 
	 * @param text
	 *            The String to send across the network.
	 * @throws InterruptedException
	 *             When socket related I/O errors occur.
	 */
	public void sendToInterface(String message, String device, String service) {

	}
}
