/**
 * @author Kostas Lekkas -- lekkas@vtrip.net
 */
package net.vtrip.things;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import org.codehaus.jettison.json.JSONArray;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;

public class Resources {
	/**
	 * HashMap < DeviceID, [set-of-requested-service-names] >
	 */
	static final HashMap<String, HashSet<String>> requests = new HashMap<String, HashSet<String>>();
	static final HashMap<String, String> messagesToPush = new HashMap<String, String>();

	/**
	 * HashMap < ReplyKey, Reply >
	 * 
	 * ReplyKey is defined as 'deviceID-serviceName'
	 */
	static HashMap<String, SensorDataReply> replies = new HashMap<String, SensorDataReply>();
	static ScheduledThreadPoolExecutor sched = new ScheduledThreadPoolExecutor(
			1);
	static long TTL_MILLISEC = 60000;

	public static String getSensorData(String deviceID, String serviceName) {
		SensorDataReply reply;
		String replyKey = deviceID + "-" + serviceName;

		synchronized (requests) {

			if (requests.containsKey(deviceID)) {
				if (!requests.get(deviceID).contains(serviceName)) {
					requests.get(deviceID).add(serviceName);
					System.out.println("added " + serviceName + " request to "
							+ deviceID);
				}
			} else { /*
					 * Create request set for this device (this means that this
					 * is the first request to this device
					 */
				HashSet<String> serviceSet = new HashSet<String>();
				serviceSet.add(serviceName);
				requests.put(deviceID, serviceSet);
				System.out.println("Created new request set for " + deviceID);
			}

			if (!replies.containsKey(replyKey)) {
				SensorDataReply r = new SensorDataReply("");
				replies.put(replyKey, r);
			}
			reply = replies.get(replyKey);
		}

		try {
			Boolean b = reply.getNewReplyQ().poll(TTL_MILLISEC,
					TimeUnit.MILLISECONDS);
			if (b == null) {
				UnregisterService(deviceID, serviceName);
				System.out.println("Time out for: " + serviceName
						+ " -- Unregistering the service");
			} else {
				reply.getNewReplyQ().add(b);
				System.out.println("Added reply to proxy reply list: "
						+ reply.getReply());
			}
			return reply.getReply();
		} catch (InterruptedException ex) {
			System.out.println("Interrupted exception: " + ex.toString());
		}

		return "";
	}

	public static void pushDataToDevice(String deviceID, String serviceName,
			String message) {

		if (messagesToPush.containsKey(deviceID))
			messagesToPush.remove(deviceID);

		messagesToPush.put(deviceID, message);
	}

	public static void newSensorDataReply(String deviceID, String serviceName,
			String reply) {
		SensorDataReply r;
		if (requests.containsKey(deviceID))
			if (requests.get(deviceID).contains(serviceName)) {
				synchronized (requests) {
					String replyKey = deviceID + "-" + serviceName;
					if (!replies.containsKey(replyKey)) {
						System.out
								.println("We have a reply for a nonexistent(time'd out?) request.");
					} else {
						r = replies.get(replyKey);
						r.setReply(reply);
						r.getNewReplyQ().add(Boolean.TRUE);
						requests.get(deviceID).remove(serviceName);
						replies.remove(replyKey);
					}
				}
			}
	}

	public static String getPendingRequests(String deviceID) {
		JSONObject reply = new JSONObject();
		JSONArray servicerequests = new JSONArray();
		try {
			synchronized (requests) {
				if (requests.containsKey(deviceID)) {
					HashSet<String> reqset = requests.get(deviceID);
					for (String s : reqset) {
						JSONObject req = new JSONObject();
						req.put("serviceName", s);
						servicerequests.put(req);
					}
				}
			}
			
			synchronized (messagesToPush) {
				if(messagesToPush.containsKey(deviceID)){
					JSONObject req = new JSONObject();
					req.put("serviceName", "pushmessage");
					req.put("message", messagesToPush.get(deviceID));
					servicerequests.put(req);
					messagesToPush.remove(deviceID);
				}
			}
			
			reply.put("pendingRequests", servicerequests);

		} catch (JSONException e) {
			e.printStackTrace();
			return "";
		}
		return reply.toString();
	}

	public static boolean UnregisterService(String deviceID, String serviceName) {
		try {
			URL url = new URL(Static_Addresses.REGISTRY_URL + "remove_service");
			URLConnection conn = null;

			try {
				conn = url.openConnection();
				conn.setDoOutput(true);
			} catch (IOException e) {
				System.out
						.println("RestRegistryImpl.removeServiceReq: could not open connection");
				e.printStackTrace();
				return false;
			}

			BufferedWriter out = null;
			try {
				out = new BufferedWriter(new OutputStreamWriter(
						conn.getOutputStream()));
			} catch (IOException e) {
				System.out
						.println("RestRegistryImpl.removeServiceReq: could not create outputStream");
				e.printStackTrace();
				return false;
			}
			try {
				String serviceURL = Static_Addresses.PROXY_URL + deviceID + "/"
						+ serviceName;
				out.write("address=" + serviceURL);

			} catch (IOException e) {
				System.out
						.println("RestRegistryImpl.removeServiceReq: could not write ouputStream");
				e.printStackTrace();
				return false;
			}
			try {
				out.flush();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
			try {
				out.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}

			BufferedReader in = null;
			try {
				in = new BufferedReader(new InputStreamReader(
						conn.getInputStream()));
			} catch (IOException e) {
				System.out
						.println("RestRegistryImpl.removeServiceReq: could not get inputStream");
				e.printStackTrace();
				return false;
			}

			String response;
			try {
				while ((response = in.readLine()) != null) {
					System.out.println(response);
				}
			} catch (IOException e) {
				System.out
						.println("RestRegistryImpl.removeServiceReq: could not read inputStream");
				e.printStackTrace();
				return false;
			}
			try {
				in.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}

		} catch (MalformedURLException e) {
			System.out
					.println("RestRegistryImpl.removeServiceReq: URL was malformed");
			e.printStackTrace();
			return false;
		}
		return true;
	}
}
