package fr.inria.arles.thinglib.core;

import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.SensorResponse;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;
//import fr.inria.arles.thinglib.devices.SoundSensorInfo;
//import fr.inria.arles.thinglib.devices.LocationSensorInfo;

import java.io.IOException;
import java.io.StreamCorruptedException;
import java.util.Vector;
import java.util.HashMap;

import android.util.Log;

public abstract class Daemon {

	public static final int STATE_DISCOVERY_UNINITIALIZED = -1;
	public static final int STATE_DISCOVERY_IN_PROGRESS   = 1;
	public static final int STATE_DISCOVERY_DONE          = 2;

	protected int mDiscoveryState;

	/**
	 * A list of all sensors in the system.
	 */
	protected Vector<SensorInfo> mSensorInfos;

	/** 
	 * A list of all worker threads for this sensor daemon. Exactly one thread
	 * is instantiated for each active Mediator used anywhere in the
	 * system.
	 */
	protected HashMap<Integer, Worker> mWorkers;

	/**
	 * Instantiates a new Daemon.
	 */
	public Daemon() {

		mDiscoveryState = STATE_DISCOVERY_UNINITIALIZED;
		mSensorInfos = null;
		mWorkers     = new HashMap<Integer, Worker>();
	}

	/**
	 * Sends a message to a Mediator inside a client process using OS-specific
	 * inter-process communications. This must be implemented for each new platform.
	 */
	protected abstract void sendMessageToClient(int clientAddress, int messageType, byte[] message);

	/**
	 * Finds within the device all sensors that are supported. This is usually 
	 * OS-dependent, so it must be implemented for each new platform. 
	 * 
	 * If you are reimplementing this for some platform, make sure
	 * that you follow the pattern below:
	 *<code>
	 *		setDiscoveryState(STATE_DISCOVERY_IN_PROGRESS);
	 *
	 *		// Override this method and do stuff here...
	 *
	 *		setDiscoveryState(STATE_DISCOVERY_DONE);
	 *		triggerOnSensorsChanged();
	 *</code>
	 */
	public abstract void discoverSensors();	

	/**
	 * Kills this Daemon, stopping all sensing immediatelly.
	 */
	public void kill() {

		for (Worker worker : mWorkers.values()) {
			worker.kill();
		}
	}

	/**
	 * Calls the correct method for each message type, and returns the response
	 * if any (null otherwise).
	 */
	protected byte[] processMessage(int clientAddress, int messageType, byte[] messagePayload) {
		
		Object replyPayloadObj = null;
		
		try {
			switch (messageType) {
	
				case Messaging.MSG_REGISTER_CLIENT:
					replyPayloadObj = new Integer(registerClient());
					break;
	
				case Messaging.MSG_DEREGISTER_CLIENT:
					unregisterClient(clientAddress);
					break;

				//case MSG_DATA_TRANSFER:
				//case MSG_REGISTER_AUTOSTART:
				//case MSG_REGISTER_AUTOSTART_REPLY:
				//case MSG_DEREGISTER_AUTOSTART:
	
				case Messaging.MSG_GET_DISCOVERY_STATE:
					replyPayloadObj = new Integer(getDiscoveryState());
					break;
	
				case Messaging.MSG_GET_SENSOR_INFOS:
					replyPayloadObj = getSensorInfos();
					break;
	
				case Messaging.MSG_GET_SENSOR_BY_UID:
					replyPayloadObj = getSensorByUid(messagePayload);
					break;
	
				case Messaging.MSG_GET_MATCHING_SENSOR:
					replyPayloadObj = getMatchingSensor(messagePayload);
					break;
	
				case Messaging.MSG_GET_MATCHING_SENSORS:
					replyPayloadObj = getMatchingSensors(messagePayload);
					break;
	
				case Messaging.MSG_SET_OUTER_SCOPE:
					setOuterScope(clientAddress, messagePayload);
					break;
	
				case Messaging.MSG_IS_RUNNING:
					replyPayloadObj = isRunning(clientAddress);
					break;
	
				case Messaging.MSG_IS_SENSING:
					replyPayloadObj = isSensing(clientAddress);
					break;
	
				case Messaging.MSG_SENSE_IMMEDIATELY:
					try {
						senseImmediately(clientAddress, messagePayload);
					} catch (ArgumentOutOfRangeException e1) {
						//TODO Auto-generated catch block
						e1.printStackTrace();
					}
					break;
	
				case Messaging.MSG_SENSE_PERIODICALLY:
					try {
						sensePeriodically(clientAddress, messagePayload);
					} catch (ArgumentOutOfRangeException e1) {
						//TODO Auto-generated catch block
						e1.printStackTrace();
					}
					break;
	
				case Messaging.MSG_SENSE_EVENT_BASED:
					senseEventBased(clientAddress);
					break;
	
				case Messaging.MSG_STOP_SENSING:
					stopSensing(clientAddress);
					break;
	
				case Messaging.MSG_ABORT_SENSING:
					abortSensing(clientAddress);
					break;
	
				case Messaging.MSG_SET_SENSOR_ENABLED:
					setSensorEnabled(clientAddress, messagePayload);
					break;
	
				case Messaging.MSG_SET_SENSOR_DISABLED:
					setSensorDisabled(clientAddress, messagePayload);
					break;
	
				case Messaging.MSG_IS_SENSOR_ENABLED:
					replyPayloadObj = isSensorEnabled(clientAddress, messagePayload);
					break;
	
				case Messaging.MSG_GET_SENSOR_RESPONSE:
					replyPayloadObj = getSensorResponse(clientAddress, messagePayload);
					break;
	
				case Messaging.MSG_GET_SENSOR_RESPONSES:
					replyPayloadObj = getSensorResponses(clientAddress);
					break;
	
				case Messaging.MSG_GET_ENABLED_SENSOR_INFOS:
					replyPayloadObj = getEnabledSensorInfos(clientAddress);
					break;
	
				default:
					break;
			}
		}
		catch (IOException e) {
			//TODO Auto-generated catch block
			e.printStackTrace();
			// TODO HERE - some things are giving me exceptions!
		}
		catch (ClassNotFoundException e) {
			//TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		byte[] replyPayload;
		try {
			replyPayload = Messaging.toBytes(replyPayloadObj);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}		
		
		return replyPayload;
	}
	
	protected int registerClient() {

		if (mSensorInfos == null) discoverSensors();

		int clientAddress = mWorkers.size();

		Worker worker = new Worker(clientAddress, this);
		mWorkers.put(clientAddress, worker);

		return clientAddress;
	}

	protected void unregisterClient(int clientAddress) {

		mWorkers.get(clientAddress).stopSensing();
		mWorkers.get(clientAddress).abortSensing();
		mWorkers.remove(clientAddress);
	}

	/**
	 * Returns the current state of the discoverer.
	 */
	public int getDiscoveryState() {

		return mDiscoveryState;
	}
	

	protected void setDiscoveryState(int newState) {

		mDiscoveryState = newState;
		for (Worker worker : mWorkers.values()) worker.triggerOnDiscoveryStateChanged(newState);
	}
	

	protected void triggerOnSensorsChanged() {

		for (Worker worker : mWorkers.values()) worker.triggerOnSensorsChanged(mSensorInfos);
	}
	

	/**
	 * Returns all discovered SensorInfo objects.
	 *
	 * @return An iterator containing the SensorInfo's of all sensors in this
	 * system.
	 */
	public Iterable<SensorInfo> getSensorInfos() {

		return mSensorInfos;
	}

	/**
	 * Returns the SensorInfo of the first sensor that matches the given UID
	 * or null if none.
	 *
	 * @param uid The UID of the SensorInfo you are looking for
	 * @return A SensorInfo object that has the given UID, or null.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public SensorInfo getSensorByUid(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		String uid = (String) Messaging.toObject(messagePayload);

		for (SensorInfo sensorInfo : mSensorInfos) {
			if (sensorInfo.isIdentifiedBy(uid)) return sensorInfo;
		}
		return null;
	}

	/**
	 * Returns the SensorInfo of the first sensor that matches the given
	 * interface ID, or null if none.
	 *
	 * @param interfaceId The interface of the SensorInfo you are looking for.
	 * @return A SensorInfo object that has the given interface, or null.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public SensorInfo getMatchingSensor(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		String interfaceId = (String) Messaging.toObject(messagePayload);

		for (SensorInfo sensorInfo : mSensorInfos) {
			if (sensorInfo.implementsInterface(interfaceId)) return sensorInfo;
		}
		return null;
	}

	/**
	 * Returns a list of SensorInfo objects that match the given
	 * interface ID.
	 *
	 * @param interfaceId The interface of the SensorInfo you are looking for.
	 * @return An array of SensorInfo objects that have the given interface.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public SensorInfo[] getMatchingSensors(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		String interfaceId = (String) Messaging.toObject(messagePayload);
		Vector<SensorInfo> sensorInfos = new Vector<SensorInfo>();

		for (SensorInfo sensorInfo : mSensorInfos) {
			if (sensorInfo.implementsInterface(interfaceId)) sensorInfos.add(sensorInfo);
		}

		return (SensorInfo[]) sensorInfos.toArray();
	}

	/**
	 * Installs a new sensor to the Daemon and calls the
	 * MediatorListener's onSensorsChanged() method.
	 *
	 * @param sensorInfo The SensorInfo of the sensor that will be added.
	 * @throws DuplicateDeviceException 
	 */
	protected void installSensor(SensorInfo sensorInfo) 
	throws DuplicateDeviceException {

		// If this sensor is a duplicate, throw exception
		if (mSensorInfos.contains(sensorInfo)) throw new DuplicateDeviceException();

		mSensorInfos.add(sensorInfo);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public void setOuterScope(int clientAddress, byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		OuterScopeWrapper scope = (OuterScopeWrapper) Messaging.toObject(messagePayload);

		mWorkers.get(clientAddress).setOuterScope(scope);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public boolean isRunning(int clientAddress) {

		return mWorkers.get(clientAddress).isRunning();
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public boolean isSensing(int clientAddress) {

		return mWorkers.get(clientAddress).isSensing();
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public void senseImmediately(int clientAddress, byte[] messagePayload) 
	throws ArgumentOutOfRangeException, StreamCorruptedException, IOException, ClassNotFoundException {

		long timeout = (Long) Messaging.toObject(messagePayload);
		mWorkers.get(clientAddress).senseImmediately(timeout);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public void sensePeriodically(int clientAddress, byte[] messagePayload)
	throws ArgumentOutOfRangeException, StreamCorruptedException, IOException, ClassNotFoundException {

		long interval = -1;
		long timeout = -1;
		
		Object[] objects = (Object[]) Messaging.toObject(messagePayload);
		try {
			interval = (long)(Long) objects[0];
			timeout = (long)(Long) objects[1];
		}
		catch (Exception e) {
			return;
		}

		mWorkers.get(clientAddress).sensePeriodically(interval, timeout);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public void senseEventBased(int clientAddress) {

		mWorkers.get(clientAddress).senseEventBased();
	}


	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public void stopSensing(int clientAddress) {
        if(mWorkers.size()>0) {
            mWorkers.get(clientAddress).stopSensing();
        }
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public void abortSensing(int clientAddress) {

		mWorkers.get(clientAddress).abortSensing();
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public void setSensorEnabled(int clientAddress, byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		SensorInfo sensorInfo = (SensorInfo) Messaging.toObject(messagePayload);
		mWorkers.get(clientAddress).setSensorEnabled(sensorInfo);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public void setSensorDisabled(int clientAddress, byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		SensorInfo sensorInfo = (SensorInfo) Messaging.toObject(messagePayload);
		mWorkers.get(clientAddress).setSensorDisabled(sensorInfo);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public boolean isSensorEnabled(int clientAddress, byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		SensorInfo sensorInfo = (SensorInfo) Messaging.toObject(messagePayload);
		return mWorkers.get(clientAddress).isSensorEnabled(sensorInfo);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	public SensorResponse getSensorResponse(int clientAddress, byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		SensorInfo sensorInfo = (SensorInfo) Messaging.toObject(messagePayload);
		return mWorkers.get(clientAddress).getSensorResponse(sensorInfo);
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public Iterable<SensorResponse> getSensorResponses(int clientAddress) {

		return mWorkers.get(clientAddress).getSensorResponses();
	}

	/**
	 * See documentation for the same-name method in Worker. The only difference
	 * is that this method also asks for the clientAddress, obtained when the
	 * client first registered with this daemon.
	 */
	public Iterable<SensorInfo> getEnabledSensorInfos(int clientAddress) {

		return mWorkers.get(clientAddress).getEnabledSensorInfos();
	}
}
