package fr.inria.arles.thinglib.core;

import java.io.IOException;
import java.io.StreamCorruptedException;

import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.SensorResponse;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;

public abstract class Mediator {

	public static final int STATE_DISCONNECTED          = -1;
	public static final int STATE_UNINITIALIZED         = 0;
	public static final int STATE_CONNECTED             = 1;
	
	protected byte[] mReplyPayload = null;
	protected int mReplyType = -1;

	protected int mConnectionState;
	protected int mSensingState;

	/**
	 * An OuterScopeWrapper object that is passed to each Sensor instance when
	 * initialized. This allows OS-specific information to be passed through the
	 * Worker in an OS-agnostic way.
	 */
	protected OuterScopeWrapper mOuterScopeWrapper;


	/**
	 * A handle representing this client when talking to the Daemon.
	 */
	protected int mClientAddress = Messaging.UNREGISTERED_ADDRESS;

	protected MediatorListener mListener = null;


	/**
	 * Calls OS-specific methods to initialize the Daemon process. If the Daemon
	 * is already running, this *does not* start a new one. Note that you
	 * usually do not need to run this method yourself, since the Daemon is
	 * initialized on-the-fly when you initialize Mediator or call connect().
	 */
	public abstract void startDaemon()
	throws DaemonConnectionException;


	/** 
	 * Calls OS-specific methods to kill the Daemon process. Note: This kills
	 * the Daemon even if there are other clients connected! To disconnect from
	 * a Daemon without killing it (and therefore ruining the fun for
	 * everybody), use the disconnect() method instead.
	 */
	public abstract void stopDaemon()
	throws DaemonConnectionException;	


	/**
	 * Uses OS-specific method calls to find out if the Daemon process is
	 * running.
	 *
	 * @return 
	 */
	public abstract boolean isDaemonRunning();

	
	/**
	 * Connects to an existing Daemon or, if necessary, launches a Daemon and
	 * then connects to it. This method *must* be called everytime you
	 * instantiate a new Mediator. The reason why this method was not added to
	 * the constructed itself is because we would like to allow different
	 * Mediator implementations to override it (and calling non-final methods
	 * from a constructor is, in general, a bad idea).
	 *
	 * Must call setConnectionState(STATE_CONNECTED) at some point; And if it 
	 * fails, the state must be set to setConnectionState(STATE_DISCOVERY_UNINITIALIZED).
	 */
	public abstract void connect()
	throws DaemonConnectionException;


	/**
	 * This method must be called at some point before the application closes,
	 * or it may lead to a memory leak in the Daemon. (We do have a stopgap
	 * measure against this kind of leak elsewhere in the code, but it's better
	 * to be safe than sorry.)
	 *
	 * Must call setConnectionState(STATE_DISCONNECTED) at some point.
	 */
	public abstract void disconnect();
	
	
	/**
	 * Sends a message to the Daemon process using OS-specific method calls.
	 *
	 * @param messageType
	 * @param messagePayload
	 */
	protected abstract void sendMessageToDaemon(int messageType, byte[] messagePayload)
	throws DaemonConnectionException;


	/**
	 * Sends a message to the Daemon and waits for a reply.
	 *
	 * @param clientAddress
	 * @param messageType
	 * @param message
	 *
	 * @return
	 * @throws BadReplyException 
	 */
	protected abstract byte[] sendRpcToDaemon(int messageType, byte[] messagePayload) 
	throws BadReplyException, DaemonConnectionException;

	
	/**
	 * Instantiates a new Mediator, but does not connect to the Daemon yet.
	 */
	public Mediator() {

		this(null);
	}


	/**
	 * Instantiates a new Mediator, but does not connect to the Daemon yet.
	 *
	 * @param listener
	 */
	public Mediator(MediatorListener listener) {

		mListener = listener;
		mConnectionState = STATE_UNINITIALIZED;
	}


	/**
	 * Called when the GC collects this instance. Here we make sure we
	 * disconnect from the Daemon if necessary.
	 */
	protected void finalize() {

		// Try to close our connection with the Daemon at this point, if
		// we haven't done that yet.
		if (mClientAddress >= 0) disconnect();
	}
	

	/**
	 * Returns the current state of the mediator-daemon connection.
	 *
	 * @return
	 */
	public int getConnectionState() {
		return mConnectionState;
	}


	/**
	 * Locally stores the state of the mediator-daemon connection, and alerts
	 * the client about the state update.
	 */
	protected void setConnectionState(int newState) {
		mConnectionState = newState;
		mListener.onConnectionStateChanged(mConnectionState);
	}


	/**
	 * This method is the callback for when the Daemon's discovery state changed. 
	 *
	 * @param message
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	protected void onDiscoveryStateChanged(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		if (mListener == null) return;

		// TODO handle exception
		
		mListener.onDiscoveryStateChanged((Integer) Messaging.toObject(messagePayload));
	}


	/**
	 * This method is the callback for when the sensing state changed. 
	 *
	 * @param message
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	protected void onSensingStateChanged(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		if (mListener == null) return;

		// TODO handle exception

		setSensingState((int)(Integer) Messaging.toObject(messagePayload));
	}


	/**
	 * Returns the current sensing state.
	 *
	 * @return
	 */
	public int getSensingState() {
		return mSensingState;
	}


	/**
	 * Locally stores the sensing state, and alerts the client about the 
	 * state update.
	 */
	protected void setSensingState(int newState) {
		mSensingState = newState;
		mListener.onSensingStateChanged(mSensingState);
	}


	/**
	 * This method is the callback for when a sensor has been added/removed.
	 *
	 * @param message
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	@SuppressWarnings("unchecked")
	protected void onSensorsChanged(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		if (mListener == null) return;

		// TODO handle exception

		Iterable<SensorInfo> sensors = (Iterable<SensorInfo>) Messaging.toObject(messagePayload);
		mListener.onSensorsChanged(sensors);
	}


	/**
	 * This method is the callback for when a sensor reports new data.
	 *
	 * @param message
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws StreamCorruptedException 
	 */
	@SuppressWarnings("unchecked")
	protected void onNewData(byte[] messagePayload) 
	throws StreamCorruptedException, IOException, ClassNotFoundException {

		if (mListener == null) return;

		// TODO handle exception

		Iterable<SensorResponse> sensorResponses = (Iterable<SensorResponse>) Messaging.toObject(messagePayload);
		mListener.onNewData(sensorResponses);
	}


	/**
	 * This is the message handler for the Mediator. It receives messages from
	 * the Daemon and demultiplexes them into Mediator methods.
	 */
	protected boolean processMessage(int messageType, byte[] messagePayload) {

		boolean processed = true;

		try {
			switch (messageType) {
	
				case Messaging.MSG_NEW_DATA:
					onNewData(messagePayload);
					break;
	
				case Messaging.MSG_DISCOVERY_STATE_CHANGED:
					onDiscoveryStateChanged(messagePayload);
					break;
	
				case Messaging.MSG_SENSING_STATE_CHANGED:
					onSensingStateChanged(messagePayload);
					break;
	
				case Messaging.MSG_SENSORS_CHANGED:
					onSensorsChanged(messagePayload);
					break;
	
				case Messaging.MSG_NEW_COMMAND:
					// nothing for now...
					break;
	
				// All *_REPLY messages defined in Messaging go here
				case Messaging.MSG_REGISTER_CLIENT_REPLY:
				case Messaging.MSG_REGISTER_AUTOSTART_REPLY:
				case Messaging.MSG_GET_DISCOVERY_STATE_REPLY:
				case Messaging.MSG_GET_SENSOR_INFOS_REPLY:
				case Messaging.MSG_GET_MATCHING_SENSOR_REPLY:
				case Messaging.MSG_GET_MATCHING_SENSORS_REPLY:
				case Messaging.MSG_IS_RUNNING_REPLY:
				case Messaging.MSG_IS_SENSING_REPLY:
				case Messaging.MSG_IS_SENSOR_ENABLED_REPLY:
				case Messaging.MSG_GET_SENSOR_RESPONSE_REPLY:
				case Messaging.MSG_GET_SENSOR_RESPONSES_REPLY:
				case Messaging.MSG_GET_ENABLED_SENSOR_INFOS_REPLY:
	
					processed = false;
					break;
	
				default:
					// TODO IMPORTANT throw exception!
					break;
			}
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (StreamCorruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return processed;
	}

	/**
	 * Sends an RPC to the Daemon, converting its reply into an Object.
	 * 
	 * @param messageType Type of RPC (i.e. one of the integer constants defined in the Messaging class).
	 * @param messagePayload An object that was serialized into a byte array using Java serialization.
	 * @return Object returned by the Daemon
	 * @throws BadReplyException
	 * @throws DaemonConnectionException 
	 */
	protected final Object getObjFromDaemon(int messageType, byte[] messagePayload) 
	throws BadReplyException, DaemonConnectionException {

		byte[] replyPayload = sendRpcToDaemon(messageType, messagePayload);

		try {
			return Messaging.toObject(replyPayload);
		} catch (StreamCorruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}
	
	/**
	 * Returns the current state of the discoverer.
	 *
	 * @return
	 * @throws BadReplyException 
	 * @throws DaemonConnectionException 
	 */
	public int getDiscoveryState() 
	throws BadReplyException, DaemonConnectionException {

		return (int) ((Integer) getObjFromDaemon(Messaging.MSG_GET_DISCOVERY_STATE, null));
	}

	/**
	 * Finds within the device all sensors that are supported.
	 * @throws DaemonConnectionException 
	 */
	public void discoverSensors() throws DaemonConnectionException {

		sendMessageToDaemon(Messaging.MSG_DISCOVER_SENSORS, null);
	}

	/**
	 * Returns all discovered SensorInfo objects.
	 *
	 * @return An iterator containing the SensorInfo's of all sensors in this
	 * system.
	 * @throws BadReplyException 
	 * @throws DaemonConnectionException 
	 */
	@SuppressWarnings("unchecked")
	public Iterable<SensorInfo> getSensorInfos() 
	throws BadReplyException, DaemonConnectionException {

		return (Iterable<SensorInfo>) getObjFromDaemon(Messaging.MSG_GET_SENSOR_INFOS, null);
	}

	/**
	 * Returns the SensorInfo of the sensor that matches the given UID, 
	 * or null if none.
	 *
	 * @param uid The UID of the SensorInfo you are looking for.
	 * @return A SensorInfo object that has the given semantic interface, or null.
	 */
	public SensorInfo getSensorByUid(String uid) {

		try {
			return (SensorInfo) getObjFromDaemon(Messaging.MSG_GET_SENSOR_BY_UID, Messaging.toBytes(uid));
		}
		catch (Exception e) {
			// TODO: handle this
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * Returns the SensorInfo of the first sensor that matches the given
	 * semantic interface, or null if none.
	 *
	 * @param interfaceId The interface name of the SensorInfo you are looking for.
	 * @return A SensorInfo object that has the given interface, or null.
	 */
	public SensorInfo getMatchingSensor(String interfaceId) {

		try {
			return (SensorInfo) getObjFromDaemon(Messaging.MSG_GET_MATCHING_SENSOR, Messaging.toBytes(interfaceId));
		}
		catch (Exception e) {
			// TODO: handle this
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * Returns a list of SensorInfo objects that match the given
	 * semantic interface.
	 *
	 * @param interfaceId The interface name of the SensorInfo you are looking for.
	 * @return An array of SensorInfo objects that has the given interface.
	 */
	public SensorInfo[] getMatchingSensors(String interfaceId) {
	// TODO MEETING: why not return an iterable?

		try {
			return (SensorInfo[]) getObjFromDaemon(Messaging.MSG_GET_MATCHING_SENSORS, Messaging.toBytes(interfaceId));
		}
		catch (Exception e) {
			// TODO: handle this
			return null;
		}
	}

	/**
	 * Passes an object of your choice (wrapped in an OuterScopeWrapper) to all sensor drivers.
	 * This is useful for OS platforms where some external "context" or "outer scope" variables
	 * are required by the sensor drivers. 
	 * 
	 * @param outerScope
	 */
	public void setOuterScope(OuterScopeWrapper outerScope) {

		try {
			sendMessageToDaemon(Messaging.MSG_SET_OUTER_SCOPE, Messaging.toBytes(outerScope));
		}
		catch (Exception e) {
			// TODO: handle this
		}		
	}

	/**
	 * Returns true if the mediator is connected
	 * 
	 * @return
	 * @throws BadReplyException
	 * @throws DaemonConnectionException 
	 */
	public boolean isRunning() 
	throws BadReplyException, DaemonConnectionException {

		if (mConnectionState == STATE_CONNECTED) {
			return (boolean) ((Boolean) getObjFromDaemon(Messaging.MSG_IS_RUNNING, null));
		}

		return false;
	}

	public boolean isSensing() 
	throws BadReplyException, DaemonConnectionException {

		return (boolean) ((Boolean) getObjFromDaemon(Messaging.MSG_IS_SENSING, null));
	}


	public void senseImmediately(long timeout) 
	throws ArgumentOutOfRangeException {

		try {
			sendMessageToDaemon(Messaging.MSG_SENSE_IMMEDIATELY, Messaging.toBytes((Long) timeout));
		}
		catch (Exception e) {
			// TODO: handle this
			e.printStackTrace();
		}
	}


	public void sensePeriodically(long interval, long timeout) 
	throws ArgumentOutOfRangeException {

		Object[] objects = new Object[2];
		objects[0] = (Object)(Long) interval;
		objects[1] = (Object)(Long) timeout;

		try {
			sendMessageToDaemon(Messaging.MSG_SENSE_PERIODICALLY, Messaging.toBytes(objects));
		}
		catch (Exception e) {
			// TODO: handle this
			e.printStackTrace();
		}
	}

	public void senseEventBased() {

		try {
			sendMessageToDaemon(Messaging.MSG_SENSE_EVENT_BASED, null);
		}
		catch (Exception e) {
			// TODO: handle this
			e.printStackTrace();
		}
	}


	public void stopSensing() 
	throws DaemonConnectionException {

		sendMessageToDaemon(Messaging.MSG_STOP_SENSING, null);
	}

	public void abortSensing() 
	throws DaemonConnectionException {

		sendMessageToDaemon(Messaging.MSG_ABORT_SENSING, null);
	}

	/**
	 * Enables/disables a sensor according to a boolean parameter. This is 
	 * just a convenience method that calls setSensorEnabled/setSensorDisabled
	 * accordingly.
	 *
	 * @param sensorInfo
	 * @param state Set to true if the sensor should be enabled, and false otherwise.
	 */
	public void setSensorEnabled(SensorInfo sensorInfo, boolean state) {

		if (state) setSensorEnabled(sensorInfo);
		else setSensorDisabled(sensorInfo);
	}

	public void setSensorEnabled(SensorInfo sensorInfo) {

		try {
			sendMessageToDaemon(Messaging.MSG_SET_SENSOR_ENABLED, Messaging.toBytes(sensorInfo));
		}
		catch (Exception e) {
			// TODO: handle this
			e.printStackTrace();
		}		
	}

	public void setSensorDisabled(SensorInfo sensorInfo) {

		try {
			sendMessageToDaemon(Messaging.MSG_SET_SENSOR_DISABLED, Messaging.toBytes(sensorInfo));
		}
		catch (Exception e) {
			// TODO: handle this
		}
	}

	public boolean isSensorEnabled(SensorInfo sensorInfo) {

		try {
			return (boolean) ((Boolean) getObjFromDaemon(Messaging.MSG_IS_SENSOR_ENABLED, Messaging.toBytes(sensorInfo)));
		}
		catch (Exception e) {
			// TODO: handle this
			return false;
		}	
	}
		

	public SensorResponse getSensorResponse(SensorInfo sensorInfo) 
	throws DeviceUnavailableException {

		try {
			return (SensorResponse) getObjFromDaemon(Messaging.MSG_GET_SENSOR_RESPONSE, Messaging.toBytes(sensorInfo));
		}
		catch (Exception e) {
			// TODO: handle this
			return null;
		}		
	}

	@SuppressWarnings("unchecked")
	public Iterable<SensorResponse> getSensorResponses() {

		try {
			return (Iterable<SensorResponse>) getObjFromDaemon(Messaging.MSG_GET_SENSOR_RESPONSES, null);
		}
		catch (Exception e) {
			// TODO: handle this
			return null;
		}
	}

	@SuppressWarnings("unchecked")
	public Iterable<SensorInfo> getEnabledSensorInfos() 
	throws BadReplyException, DaemonConnectionException {

		return (Iterable<SensorInfo>) getObjFromDaemon(Messaging.MSG_GET_ENABLED_SENSOR_INFOS, null);
	}
}

