package fr.inria.arles.thinglib.core;

import fr.inria.arles.thinglib.devices.Sensor;
import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.SensorListener;
import fr.inria.arles.thinglib.devices.SensorResponse;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import android.util.Log;

public class Worker {
	
	/** 
	 * Describes the Worker's current state as "closed" through the kill() method.
	 */

	public static final int STATE_CLOSED = 0;

	/** 
	 * Describes the Worker's current state as idle (not sensing, and is not
	 * waiting for any timers).
	 */
	public static final int STATE_STOPPED = 1;

	/**
	 * Describes the Worker's current state as being in the process of stopping.
	 */
	public static final int STATE_STOPPING = 2;
 
	/**
	 * Describes the Worker's current state as waiting for the next sampling
	 * period, or waiting for events (in case of event-based sensors).
	 */
	public static final int STATE_WAITING_FOR_TIMER_OR_EVENT = 3;

	/** 
	 * Describes the Worker's current state as actually sampling data from the
	 * sensors.
	 */
	public static final int STATE_SENSING = 4;

	/**
	 * A list of enabled Sensors.
	 */
	protected HashMap<SensorInfo, Sensor> mEnabledSensors;

	/**
	 * A hashmap that tells us whether a given sensor has already responsed. (In
	 * reality, we would love it if we could just create a HashMap<SensorInfo,Sensor>
	 * that would replace this one and the mSensorResponses list. However, this
	 * leads to problems later on when calling triggerOnNewData, since HashMap's
	 * values() method does not return a serializable object. So we'd have to
	 * copy it into a serializable, leading to performance issues.)
	 */
	protected HashMap<SensorInfo, Boolean> mSensorsWithResponses;

	protected Vector<SensorResponse> mSensorResponses;
	
	/**
	 * An OuterScopeWrapper object that is passed to each Sensor instance when
	 * initialized. This allows OS-specific information to be passed through the
	 * Worker in an OS-agnostic way.
	 */
	protected OuterScopeWrapper mOuterScopeWrapper;

	protected Daemon mDaemon;
	protected int mClientHandle;

	/**
	 * This variable holds the current state of the controller, which behaves as
	 * a state machine. This variable is never directly changed by the user's
	 * actions in the GUI. It only changes whenever it is necessary in order to
	 * reflect the controller's *actual* state.
	 */
	protected int mState;

	/**
	 * This variable holds the next state of the controller. 
	 */
	protected int mRequestedState;

	/**
	 * mExecutor works as a timer, calling the sensing method every few minutes.
	 */
	protected final ScheduledExecutorService mExecutor = Executors.newSingleThreadScheduledExecutor();

	/**
	 * These are objects used to cancel tasks run by the mExecutor.
	 */
	protected ScheduledFuture<?> mScheduledSensingTask, mScheduledTimeoutTask;


	/**
	 * Instantiates a new Worker.
	 */
	public Worker(int clientHandle, Daemon daemon) {

		mState          = STATE_STOPPED;
		mDaemon         = daemon;
		mClientHandle   = clientHandle;
		mEnabledSensors = new HashMap<SensorInfo, Sensor>();
	}

	public void triggerOnDiscoveryStateChanged(int state) {

		// TODO handle exception

		try {
			mDaemon.sendMessageToClient(mClientHandle, Messaging.MSG_DISCOVERY_STATE_CHANGED, Messaging.toBytes((Integer) state));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}	

	public void triggerOnSensingStateChanged(int state) {

		// TODO handle exception

		try {
			mDaemon.sendMessageToClient(mClientHandle, Messaging.MSG_SENSING_STATE_CHANGED, Messaging.toBytes((Integer) state));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}	

	public void triggerOnSensorsChanged(Iterable<SensorInfo> sensors) {

		// TODO handle exception

		try {
			mDaemon.sendMessageToClient(mClientHandle, Messaging.MSG_SENSORS_CHANGED, Messaging.toBytes(sensors));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void triggerOnNewData(Iterable<SensorResponse> sr) {
	
		// TODO handle exception

		try {
			mDaemon.sendMessageToClient(mClientHandle, Messaging.MSG_NEW_DATA, Messaging.toBytes(sr));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace(); // ERROR: not serializable (hashmap values)
		}
	}

	public void kill() {

		abortSensing();
		mState = STATE_CLOSED;
	}

	/**
	 * Sets the OuterScopeWrapper instance that will be passed onto the sensors. This
	 * way you can pass onto the sensor drivers any OS-dependent objects that
	 * are accessible from the scope there this method is called. For instance,
	 * on Android it is a good idea to pass the local Context, wrapped in an
	 * OuterScopeWrapper object.
	 */
	public void setOuterScope(OuterScopeWrapper scope) {

		mOuterScopeWrapper = scope;
	}

	/**
	 * Returns true if the worker is either sensing or waiting for the next
	 * sensing task to be triggered.
	 */
	public boolean isRunning() {

		return (mState == STATE_WAITING_FOR_TIMER_OR_EVENT || mState == STATE_SENSING);
	}

	/**
	 * Returns true if the worker is currently sensing any data, but false
	 * if it is either stopped or waiting for the next sensing task to be triggered.
	 */
	public boolean isSensing() {

		return (mState == STATE_SENSING);
	}

	/** 
	 * Immediatelly starts the threads to ask all enabled sensors for readings. 
	 * After all sensors respond, the MediatorListener's onNewData() method is
	 * called.
	 *
	 * @param timeout Sets the maximum time (in milliseconds) to wait for a sensor to
	 * respond, which applies only to sensors that are not event-driven. After
	 * this timeout, the sensor will be cancelled. If this case, only the
	 * responses of the sensors that have not timed out will be reported through
	 * the onNewData() method.
	 *
	 * @throws ArgumentOutOfRangeException
	 */
	public void senseImmediately(long timeout) 
	throws ArgumentOutOfRangeException {

		if (timeout <= 0) throw new ArgumentOutOfRangeException("Timeout must be greater than zero.");

		if (isRunning()) return; // TODO: Throw exception?
		if (mEnabledSensors.isEmpty()) return;		

		mRequestedState = STATE_SENSING;

		mScheduledTimeoutTask = mExecutor.schedule(mTimeoutTask, timeout, TimeUnit.MILLISECONDS);
		mExecutor.submit(mSensingTask);
	}

	/** 
	 * Immediatelly starts the threads to ask all enabled sensors for readings,
	 * and asks for new readings periodically. After all sensors respond, the
	 * MediatorListener's onNewData() method is called.
	 *	
	 * @param interval Sets the sensing interval (in milliseconds), which is only
	 * used for sensors that are not event-driven. Note that this is the sensing
	 * interval between two consecutive *requests* for a reading.  The actual
	 * sensing period between readings consists of this interval plus the
	 * internal delays of the sensor and sensor driver. Also note that this is
	 * not a *sampling* period, as a sensor driver may use several samples per
	 * reading, depending on the implementation of the Sensor driver (e.g.  a
	 * driver could sample 10 times very quickly and return the average value).
	 *
	 * @param timeout Sets the maximum time (in milliseconds) to wait for a sensor to
	 * respond, which applies only to sensors that are not event-driven. After
	 * this timeout, the sensor will be cancelled. If this case, only the
	 * responses of the sensors that have not timed out will be reported through
	 * the onNewData() method.
	 *
	 * @throws ArgumentOutOfRangeException
	 */
	public void sensePeriodically(long interval, long timeout) 
	throws ArgumentOutOfRangeException {

		if (interval <= 0) throw new ArgumentOutOfRangeException("Interval must be greater than zero.");
		if (timeout  <= 0) throw new ArgumentOutOfRangeException("Timeout must be greater than zero.");

		if (isRunning()) return; // TODO: Throw exception?
		if (mEnabledSensors.isEmpty()) return;

		mRequestedState = STATE_SENSING;
		
		mScheduledTimeoutTask = mExecutor.scheduleAtFixedRate(mTimeoutTask, timeout, interval, TimeUnit.MILLISECONDS);
		mScheduledSensingTask = mExecutor.scheduleAtFixedRate(mSensingTask, 0      , interval, TimeUnit.MILLISECONDS);
	}

	/** 
	 * Immediatelly tells all event-based sensors to start listening for new
	 * events. Whenever any sensor responds, the MediatorListener's onNewData()
	 * method is called.
	 */
	public void senseEventBased() {

		if (isRunning()) return; // TODO: Throw exception?
		if (mEnabledSensors.isEmpty()) return;

		mRequestedState = STATE_WAITING_FOR_TIMER_OR_EVENT;
		setState(STATE_WAITING_FOR_TIMER_OR_EVENT);

		Sensor sensor;
		SensorInfo sensorInfo;

		// go through every enabled event-based sensor and call its getData method
		for (Map.Entry<SensorInfo, Sensor> entry: mEnabledSensors.entrySet()) {

			sensor = entry.getValue();
			sensorInfo = entry.getKey();

			if (sensorInfo.isEventBased()) {
				sensor.getData(mDataHandlerEventBased); 
			}
		}
	}


	/**
	 * Politely stops the sensing timer. That is, if there currently is a sensor
	 * task in execution, this method returns without aborting it (although it
	 * does cancel all subsequent sensing tasks).
	 */
	public void stopSensing() {

		// make sure we don't stop sensing unless it is already running
		if (!isRunning()) return;

		setState(STATE_STOPPING);
		mRequestedState = STATE_STOPPED;

		if (mScheduledTimeoutTask != null) mScheduledTimeoutTask.cancel(false);
		if (mScheduledSensingTask != null) mScheduledSensingTask.cancel(false);

		setState(STATE_STOPPED);
	}

	/**
	 * Stops the sensing timer *and* aborts the current sensing task, if any.
	 */
	public void abortSensing() {

		// make sure we don't stop sensing unless it is already running
		if (!isRunning()) return;

		setState(STATE_STOPPING);
		mRequestedState = STATE_STOPPED;

		if (mScheduledTimeoutTask != null) mScheduledTimeoutTask.cancel(true);
		if (mScheduledSensingTask != null) mScheduledSensingTask.cancel(true);

		for (Sensor sensor : mEnabledSensors.values()) {
			try { 
				sensor.cancelCurrentTask(); 
			}
			catch (Exception e) { 
				// TODO: log exception
				e.printStackTrace();
			}
		}

		setState(STATE_STOPPED);
	}

	/**
	 * Enable a given sensor.
	 *
	 * @param sensorInfo A SensorInfo describing the sensor in question.
	 */
	public void setSensorEnabled(SensorInfo sensorInfo) { 

		if (sensorInfo == null) return; // TODO: exception?

		Sensor sensor = mEnabledSensors.get(sensorInfo);

		// don't add duplicate sensors!
		if (sensor != null) return;

		sensor = sensorInfo.getSensorInstance(mOuterScopeWrapper);
		mEnabledSensors.put(sensorInfo, sensor);
	}

	/**
	 * Disable a given sensor.
	 *
	 * @param sensorInfo A SensorInfo describing the sensor in question.
	 */
	public void setSensorDisabled(SensorInfo sensorInfo) { 

		if (sensorInfo == null) return; // TODO: exception?

		mEnabledSensors.remove(sensorInfo);
	}

	/**
	 * Returns true if the sensor is enabled, false otherwise.
	 *
	 * @param sensorInfo A SensorInfo describing the sensor in question.
	 * @return A boolean containing the state of the sensor (enabled = true,
	 * disabled = false)
	 */
	public boolean isSensorEnabled(SensorInfo sensorInfo) { 

		return mEnabledSensors.containsKey(sensorInfo);
	}

	/**
	 * Gets the latest SensorResponse related to a given sensor.
	 *
	 * @param sensorInfo A SensorInfo describing the sensor in question.
	 * @throws DeviceUnavailableException Thrown if the sensor defined by
	 * sensorInfo is not enabled or does not exist in the system.
	 */
	public SensorResponse getSensorResponse(SensorInfo sensorInfo) 
	throws DeviceUnavailableException { 

		Sensor sensor = mEnabledSensors.get(sensorInfo);
		if (sensor == null) throw new DeviceUnavailableException();

		return sensor.getData();
	}

	/**
	 * Gets the latest SensorResponses for all enabled sensors. If a sensor is
	 * unavailable for some reason (for instance, if sensing failed) it is
	 * simply not included in the return.
	 *
	 * @return An Iterable containing the latest SensorResponse objects for all
	 * enabled sensors.
	 */
	public Iterable<SensorResponse> getSensorResponses() {

		return mSensorResponses;
	}

	/**
	 * Gets the SensorInfo objects of all enabled sensors.
	 *
	 * @return An iterator containing the SensorInfos of all enabled sensors in
	 * this system.
	 */
	public Iterable<SensorInfo> getEnabledSensorInfos() {

		return mEnabledSensors.keySet();
	}

	/**
	 * Change the state of the internal state machine and alert the outside
	 * world about the state change.
	 */
	protected void setState(int state) {

		mState = state;
		triggerOnSensingStateChanged(mState);
	}

	/**
	 * Forward the sensor output to the DataListeners
	 */
	protected SensorListener mDataHandler = new SensorListener() {

		@Override
		public void onNewResponse(SensorResponse response) {

			// if this response is not a duplicate, add it to our response list
			if (!mSensorsWithResponses.containsKey(response.sensorInfo) || 
					!mSensorsWithResponses.get(response.sensorInfo)) {

				mSensorsWithResponses.put(response.sensorInfo, true);
				mSensorResponses.add(response);
				// if we have all responses, inform the client
				if (mSensorResponses.size() >= mEnabledSensors.size()) {

					if (mRequestedState != STATE_STOPPED) 
						setState(STATE_WAITING_FOR_TIMER_OR_EVENT);

					triggerOnNewData(mSensorResponses);
				}
			}
		}
	};

	/**
	 * Forward the sensor output to the DataListeners
	 */
	protected SensorListener mDataHandlerEventBased = new SensorListener() {

		@Override
		public void onNewResponse(SensorResponse response) {

			// wrap the response in an Iterable (in this case a Vector) so it's
			// compatible with the onNewData method. 
			//
			// (It would be nice to find a very lightweight Iterable for use
			// here, but I guess Vector will do for now)
			// 
			Vector<SensorResponse> wrappedResponse = new Vector<SensorResponse>(1);
			wrappedResponse.add(response);

			triggerOnNewData(wrappedResponse);
		}
	};

	/**
	 * When executed, this tells all sensors to capture new data.
	 */
	protected final Runnable mSensingTask = new Runnable() {

		@Override
		public void run() {

			setState(STATE_SENSING);

			resetSensorResponses();

			Sensor sensor;
			SensorInfo sensorInfo;

			// ask all enabled non-event-based sensors to fetch a value
			for (Map.Entry<SensorInfo, Sensor> entry: mEnabledSensors.entrySet()) {

				sensor = entry.getValue();
				sensorInfo = entry.getKey();

				if (!sensorInfo.isEventBased()) {
					sensor.getData(mDataHandler); 
				}
			}
		}
	};
	
	/**
	 * When executed, this cancels all running sensors and reports to the client
	 * the data that had been collected up to here.
	 */
	protected final Runnable mTimeoutTask = new Runnable() {

		@Override
		public void run() {
			
			Sensor sensor;
			SensorInfo sensorInfo;
			
			for (Map.Entry<SensorInfo, Sensor> entry: mEnabledSensors.entrySet()) {

				sensor = entry.getValue();
				sensorInfo = entry.getKey();
				
				// make sure we do not cancel event-based sensors
				if (sensorInfo.isEventBased()) continue;

				if (mSensorsWithResponses.containsKey(sensorInfo)) continue;
				
				try {
					sensor.cancelCurrentTask(); 

					// create an empty response for this sensor

					SensorResponse response = new SensorResponse(
							sensor.getSensorInfo(), SensorResponse.TIMEOUT);

					mSensorsWithResponses.put(sensorInfo, true);
					mSensorResponses.add(response);
				}
				catch (Exception e) {
					// TODO: log exception	   	
					e.printStackTrace();
				}

			}

			if (mRequestedState != STATE_STOPPED) 
				setState(STATE_WAITING_FOR_TIMER_OR_EVENT);

			triggerOnNewData(mSensorResponses);
		}
	};

	protected void resetSensorResponses() {

		if (mSensorResponses == null || mSensorResponses.size() != mEnabledSensors.size()) {

			mSensorsWithResponses = new HashMap<SensorInfo, Boolean>(mEnabledSensors.size());
			mSensorResponses = new Vector<SensorResponse>(mEnabledSensors.size());
		}
		else {
			mSensorResponses.clear();
		}
		
		for (SensorInfo sensorInfo : mEnabledSensors.keySet()) {
			mSensorsWithResponses.put(sensorInfo, false);
		}
	}

}
