package fr.inria.arles.thinglib.devices;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import fr.inria.arles.thinglib.datatypes.DefaultIntegerData;

public class DummySensor implements Sensor {

	protected int mCounter;
	protected final SensorInfo mSensorInfo;
	protected int mIncrement;
	protected int mMax;
	protected final HashMap<String, Object> mParameters;
	
	public final String PARAM_MAX = "maximum";
	public final String PARAM_INCREMENT = "increment amount";

	public DummySensor(SensorInfo sensorInfo) {
		
		mCounter = 0;
		mIncrement = 1;
		mMax = 1000;
		
		mSensorInfo = sensorInfo;
		
		mParameters = new HashMap<String, Object>(2);
		mParameters.put(PARAM_MAX, (Integer) mMax);
		mParameters.put(PARAM_INCREMENT, (Integer) mIncrement);
	}

	@Override
	public SensorInfo getSensorInfo() {
		return mSensorInfo;
	}

	@Override
	public SensorResponse getData() {

		long timestamp = ((long) (System.currentTimeMillis())) * 1000000;
		
		SensorResponse response = new SensorResponse(
				mSensorInfo, new DefaultIntegerData(mCounter, timestamp));

		mCounter += mIncrement;
		if (mCounter > mMax) mCounter = 0;

		return response;
	}

	@Override
	public void getData(SensorListener handler) {
		// This sensor is not really an event-based sensor, so we just call
		// getData() and send it to the handler immediately.
		handler.onNewResponse(this.getData());
	}

	@Override
	public boolean hasFreshData() {
		return true;
	}

	@Override
	public void cancelCurrentTask() {
		// This method does nothing, since this is not an event-based sensor
		mCounter = 0;
	}

	@Override
	public boolean isSensing() {
		return false;
	}

	@Override
	public Object getParameter(String key) 
	throws ParameterErrorException {
		
		if (key.equals(PARAM_MAX) || key.equals(PARAM_INCREMENT)) {
			return mParameters.get(key);
		}	

		throw new ParameterErrorException(String.format("No such parameter: %s", key));
	}

	@Override
	public void setParameter(String key, Object valueObj)
	throws ParameterErrorException {
		
		int value;
		
		try {
			value = (Integer) valueObj;
		}
		catch (Exception e) {
			throw new ParameterErrorException(String.format("Parameter %s is not an Integer.", key));			
		}

		if (key.equals(PARAM_MAX)) {
			if (value > 0) mMax = value;
			else throw new ParameterErrorException(String.format("Parameter %s=%d is negative or zero.", key, value));
		}
		
		else if (key.equals(PARAM_INCREMENT)) {
			if (value > 0) mIncrement = value;
			else throw new ParameterErrorException(String.format("Parameter %s=%d is negative or zero.", key, value));
		}
		
		else throw new ParameterErrorException(String.format("No such parameter: %s", key));
		
		mParameters.put(key, value);		
	}

	@Override
	public Map<String, Object> getParameters() {

		return Collections.unmodifiableMap(mParameters);
	}
}


