package fr.inria.arles.thinglib.devices;

import java.util.Map;

/** 
 * Interface for sensor drivers, which are the objects that hold the actual
 * logic behind each sensor's operation (as opposed to SensorInfo objects, which
 * hold sensor metadata).
 */
public interface Sensor {

	/**
	 * Used when passing a SensorResponse, to declare that the corresponding
	 * sensing task has been successful. 
	 */
	public static final int SENSING_SUCCESS = 0;

	/**
	 * Returns a SensorInfo object that describes this sensor . 
	 */
	public abstract SensorInfo getSensorInfo();	

	/**
	 * Sets a single sensor parameter, if any are supported, by passing it as key-value
	 * pair. If the parameter is not supported, or is out of range, or any other 
	 * error takes place, a ParameterErrorException is thrown.
	 *
	 * @param key A string representing the parameter to be set.
	 * @param value The new value for the parameter.
	 */
	public abstract void setParameter(String key, Object value)
	throws ParameterErrorException;	
	
	/**
	 * Get the value of a single parameter. If the paremeter does not exist, returns null;
	 *
	 * @return A HashMap containing key-value pairs of parameter names and parameter 
	 * values.
	 */
	public abstract Object getParameter(String key)
	throws ParameterErrorException;
		
	/**
	 * Get a map with key-value pairs for all of this class's parameterers. The returned Map 
	 * should be considered immutable. If there are no parameters, then you should return either
	 * an empty Map or just null.
	 * 
	 * This method is useful when programattically listing a sensor's parameters in a GUI, 
	 * for instance.
	 *
	 * @return An immutable Map containing key-value pairs of parameter names and parameter 
	 * values.
	 */
	public abstract Map<String, Object> getParameters();
	// TODO: should this move into SensorInfo instead? (in which case it would not return the 
	// values of each parameter, of course)

	/**
	 * Asks the sensor for a (possibly old) data point. Synchronous: returns
	 * immediately, even if that means returning an old value.
	 */
	public abstract SensorResponse getData();

	/**
	 * Asks the sensor for a new datapoint. Asynchronous.
	 *
	 * @param handler A Handler object to be executed when the sensor has a
	 * new value. If this Sensor is event-based, this method starts listening
	 * for data, and calls the handler whenever new events are detected.
	 */
	public abstract void getData(SensorListener handler);

	/**
	 * Returns true if the sensor knows for sure that the data is "fresh" (i.e.
	 * valid as a good approximation of the true current status).
	 * This is loosely defined on purpose.
	 */
	public abstract boolean hasFreshData(); 

	/**
	 * Cancels the latest getData(handler). Used, for instance, in case the
	 * sensor is taking too long to return. In event-based sensors, this is used
	 * to stop listening for new events.
	 */
	public abstract void cancelCurrentTask();

	/**
	 * Returns the state of the sensor.
	 *
	 * @return True if the sensor is currently busy performing a sensing task
	 * that was initiated with getData()/getData(handler), false otherwise.
	 */
	public abstract boolean isSensing();
}

