package fr.inria.arles.thinglib.devices;

import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;
import fr.inria.arles.thinglib.semantics.SemanticClass;

import java.io.Serializable;

/**
 * This class contains metadata about a sensor. For thread-safety, objects of this 
 * class are immutable (all attributes are constant).
 */
public abstract class SensorInfo extends SemanticClass implements Serializable {

	private static final long serialVersionUID = -4344256524861465380L;

	/** 
	 * Returns the full name of this sensor, as it will be shown to the user
	 * wherever there is room. For example: "GPS-based location sensor".
	 * Required.
	 */
	public abstract String getLongName();

	/**
	 * Returns the shorter name for the sensor, to be shown in summaries and
	 * legends for plots. For example: "GPS". Required.
	 */
	public abstract String getShortName();

	/**
	 * Returns an explanation of what this sensor does, to be displayed to the
	 * end-user, who is assumed to know nothing about this sensor. This can
	 * consists of multiple sentences. Required.
	 */
	public abstract String getDescription();

	/**
	 * Returns the name of the person or company in charge of developing and 
	 * maintaining this sensor driver. Required.
	 */
	public abstract String getAuthor();

	/**
	 * Returns true if a sensor is an event-based sensor (i.e. it cannot be
	 * called synchronously, since it returns values in a stochastic manner).
	 * Typical examples of such sensors include PIR motion sensors, break-beam
	 * sensors, and so on.
	 */
	public abstract boolean isEventBased();

	/**
	 * Returns the URL where the used can get more information and support for
	 * this sensor driver. Optional: returns null if no URL.
	 */
	public String getDocumentationUrl() {return null;}

	// TODO: How do we internationalize these strings?

	/**
	 * Returns true if this SensorInfo's UID matches the given SensorInfo's UID.
	 *
	 * @param that A SensorInfo instance that will be compared to this SensorInfo.
	 */
	public boolean equals(SensorInfo that) {
		return this.getUid().equals(that.getUid());
	}

	/**
	 * Factory method to obtains a Sensor instance. This could be a new instance
	 * or a reusable one (i.e. a singleton), depending on the specific design of
	 * the sensor driver.
	 */
	public abstract Sensor getSensorInstance(OuterScopeWrapper outerScope);
	// I would like this to be static, but Java does not allow abstract static :-/
	// So this means that I have to instantiate SensorInfo objects and pass them 
	// around everywhere instead of passing a reference to the class. Oh well...
}

