package fr.inria.arles.thinglib.devices;

import fr.inria.arles.thinglib.datatypes.SensorData;
import fr.inria.arles.thinglib.semantics.SemanticClass;

import java.util.Comparator;
import java.io.Serializable;

/** 
 * A SensorResponse is the response that a sensor passes to its callee whenever
 * it is asked for some data. It carries both the sensor data itself as it does
 * some metadata about it, such as the SensorInfo of the sensor that produced
 * this response, and a status code. This object is immutable.
 */
public class SensorResponse extends SemanticClass implements Serializable {

	private static final long serialVersionUID = -778142383750028140L;

	/**
	 * Constant used to indicate that the sensor has returned data correctly. Do
	 * not redefine this value: SUCCESS must be 0. 
	 */
	public static final int SUCCESS = 0;

	/**
	 * Constant used to indicate that the measurements has timed out. Do not
	 * redefine this value.
	 */
	public static final int TIMEOUT = Integer.MIN_VALUE;

	/**
	 * Constant used to indicate that the sensor is unavailable.
	 */
	public static final int UNAVAILABLE = Integer.MIN_VALUE + 1;

	/**
	 * Constant used to indicate that the message was corrupted.
	 */
	public static final int CORRUPTED = Integer.MIN_VALUE + 2;

	/**
	 * Constant used to indicate an unknown status, such as when the
	 * SensorResponse was created not from a sensor reading, but from a record
	 * in a file, for instance. Do not redefine this value.
	 */
	public static final int UNKNOWN = Integer.MAX_VALUE;

	/**
	 * A SensorInfo object referring to the sensor that produced this response.
	 */
	public final SensorInfo sensorInfo;

	/** 
	 * An integer carrying an error or warning code if the sensing was
	 * unsuccessful or SensorResponse.SUCCESS otherwise. Errors should be
	 * represented with negative numbers, while warnings with positive numbers.
	 *
	 * Note that the intervals [Integer.MIN_VALUE, Integer.MIN_VALUE+127] and
	 * [Integer.MAX_VALUE-127, Integer.MAX_VALUE] are reserved for internal use.
	 * Your own status integers must be outside that range (which they most
	 * probably are, anyway).
	 */
	public final int status;

	/**
	 * The actual data measured by the sensor.
	 */
	public final SensorData payload;

	/**
	 * Creates a data-less SensorResponse, indicating that an error ocurred.
	 *
	 * @param sensorInfo The SensorInfo object for the sensor that produced 
	 * this response.
	 *
	 * @param status An error code.
	 */
	public SensorResponse(SensorInfo sensorInfo, int status) {
		this(sensorInfo, status, null);
	}

	/**
	 * Creates a SensorResponse for a successful data-sensing period (i.e. it
	 * sets status to Sensor.SENSING_SUCCESS)
	 *
	 * @param sensorInfo The SensorInfo object for the sensor that produced 
	 * this response.
	 *
	 * @param payload The actual data measured by the sensor.
	 */
	public SensorResponse(SensorInfo sensorInfo, SensorData payload) {
		this(sensorInfo, Sensor.SENSING_SUCCESS, payload);
	}

	/**
	 * Creates a SensorResponse instance.
	 *
	 * @param sensorInfo The SensorInfo object for the sensor that produced 
	 * this response.
	 *
	 * @param status An error status number, defined by the sensor
	 * driver (which extends Sensor) as some negative integer. If there was no
	 * error, Sensor.SENSING_SUCCESS should be passed here instead.
	 *
	 * @param payload The actual data measured by the sensor, which gets
	 * copied inside this method for thread-safety.
	 */
	public SensorResponse(SensorInfo sensorInfo, int status, SensorData payload) {

		this.sensorInfo = sensorInfo; 
		this.status     = status;
		this.payload    = (payload == null)? null : payload.copy();
	}

	/**
	 * Overrides SemanticClass's getUid() method.
	 */
	@Override
	public String getUid() {
		return "fr.inria.arles.thinglib.SensorResponse";
	}

	/**
	 * Overrides SemanticClass's getInterface() method. Note that 
	 * SensorResponse interfaces differ from the SensorData interfaces
	 * in that the former only defines properties associated with the 
	 * response itself (such as status codes) while the latter deals with 
	 * aspects of the data, such as physical units and number of components. 
	 */
	@Override
	public String getInterface() {
		return "fr.inria.arles.thinglib.SensorResponse";
	}

	/**
	 * Return true if this SensorResponse originated from the same sensor as the
	 * given SensorResponse, as determined by matching their SensorInfo objects'
	 * UID.
	 */
	public boolean hasSameSource(SensorResponse that) {
		return this.sensorInfo.equals(that.sensorInfo);
	}

	/**
	 * A comparator the lets you sort SensorResponse lists in a way that is
	 * independent of the actual response instance. For this, this comparator
	 * compares the UID strings of each SensorResponse pair, rather than
	 * comparing their address in memory.
	 *
	 * Note: this comparator imposes orderings that are inconsistent with
	 * equals.
	 */
	public static class SourceComparator implements Comparator<SensorResponse> {

		@Override
		public int compare(SensorResponse sr1, SensorResponse sr2) {
			return sr1.sensorInfo.getUid().compareTo(sr2.sensorInfo.getUid());
		}
	}

	public String toString() {
		return 
			  "Source: " + sensorInfo.getShortName() + "\n"
			+ "Status: " + Integer.toString(status) + "\n"
			+ "Payload:\n" + payload.toString();
	}
}
