package fr.inria.arles.thinglib.devices.android;

import java.util.Map;

import android.content.Context;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.text.format.Time;

import fr.inria.arles.thinglib.datatypes.DefaultDouble3dData;
import fr.inria.arles.thinglib.devices.IntervalSensor;
import fr.inria.arles.thinglib.devices.ParameterErrorException;
import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.SensorListener;
import fr.inria.arles.thinglib.devices.SensorResponse;
import fr.inria.arles.thinglib.math.InvalidDataRangeException;
import fr.inria.arles.thinglib.math.SimpleStatistics;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;

public class FallSensor implements IntervalSensor {

	/** Detect a fall whenever any of the orientation angles' standard deviation is over this value. */
	private final double FALL_THRESHOLD = 0.5; 

	/** Number of samples to use for the standard-deviation calculation. */
	private final int MAX_SAMPLES = 50; 

	SensorManager mSensorManager;
	SensorListener mSensorListener;
	boolean mIsRunning;
	private final SensorInfo mSensorInfo;

	/**
	 * Specifies how old data can be to still be considered "fresh", in
	 * milliseconds.
	 */
	long mFreshnessMargin;

	double[] mXbuff, mYbuff, mZbuff;
	long mT;

	SimpleStatistics mX, mY, mZ;

	SensorResponse mLatestSensorResponse;

	int mCurrNumOfSamples;    
	int mSensingPeriod;


	public FallSensor(OuterScopeWrapper scope, SensorInfo sensorInfo) {

		Context context = (Context) scope.getContents();

		mIsRunning = false;
		mSensorInfo = sensorInfo;
		mFreshnessMargin = 3000; // in milliseconds
		mSensorManager = (SensorManager) context.getSystemService(Context.SENSOR_SERVICE);

		mX = new SimpleStatistics();
		mY = new SimpleStatistics();
		mZ = new SimpleStatistics();
		mT = -1;

		// set sensing interval to defaults
		// (must come after initialization of mX, mY, etc.)
		setSensingInterval(-1, -1);
	}


	@Override
	public SensorResponse getData() {

		return mLatestSensorResponse;
	}

	@Override
	public void setSensingInterval(long sensingPeriod, int totalNumberOfSamples) {

		mSensingPeriod = SensorManager.SENSOR_DELAY_GAME;

		afterSetSensingInterval();
	}

	@Override
	public void getData(SensorListener handler) {

		mSensorListener = handler;
		triggerDataCollection();
	}

	@Override
	public boolean hasFreshData() {

		Time time = new Time();
		time.setToNow();

		// TODO: Make sure we're handling timezones correctly (toMillis argument)
		if (mLatestSensorResponse.payload.timestamp > time.toMillis(false) - mFreshnessMargin) return true;

		return false;
	}

	@Override
	public void cancelCurrentTask() {

		if (mIsRunning) stopDataCollection();
	}

	@Override
	public boolean isSensing() {

		return mIsRunning;
	}
	
	@Override
	public Object getParameter(String key) 
	throws ParameterErrorException {

		// This method does nothing, since this class has no parameters
		throw new ParameterErrorException("This class has no parameters!");
	}

	@Override
	public void setParameter(String key, Object value)
	throws ParameterErrorException {
		
		// This method does nothing, since this class has no parameters
		throw new ParameterErrorException("This class has no parameters!");
	}

	@Override
	public Map<String, Object> getParameters() {

		// This method does nothing, since this class has no parameters
		return null;
	}	

	@Override
	public SensorInfo getSensorInfo() {
		return mSensorInfo;
	}

		
	// ******************************************************************************
	// End of Overrides. The actual implementation of the Fall Detector is next

	private final SensorEventListener mSensorEventListener = new SensorEventListener() {

		/**
		 * This listener is called whenever there's a new inertial reading available.
		 */
		@Override
		public void onSensorChanged(SensorEvent event) {

			if (event.sensor.getType() != android.hardware.Sensor.TYPE_ORIENTATION) return;

			if (mCurrNumOfSamples < MAX_SAMPLES) {

				if (mCurrNumOfSamples == 0) mT = event.timestamp;

				mXbuff[mCurrNumOfSamples] = event.values[0];
				mYbuff[mCurrNumOfSamples] = event.values[1];
				mZbuff[mCurrNumOfSamples] = event.values[2];

				mCurrNumOfSamples++;

				if (mCurrNumOfSamples == MAX_SAMPLES) {
					//stopDataCollection();
					computeStatistics();
					if (isFall()) prepareDataAndCallHandler();
					mCurrNumOfSamples = 0;
				}
			}
		}

		@Override
		public void onAccuracyChanged(android.hardware.Sensor sensor, int accuracy) {}
	};    

	protected void afterSetSensingInterval() {
		mXbuff = new double[MAX_SAMPLES];
		mYbuff = new double[MAX_SAMPLES];
		mZbuff = new double[MAX_SAMPLES];
		mX.setData(mXbuff);
		mY.setData(mYbuff);
		mZ.setData(mZbuff);
	}

	private void triggerDataCollection() {

		mIsRunning = true;
		mCurrNumOfSamples = 0;

		boolean res = mSensorManager.registerListener(mSensorEventListener, 
		        mSensorManager.getDefaultSensor(android.hardware.Sensor.TYPE_ORIENTATION),
		        mSensingPeriod);

		// TODO IMPORTANT:
		// if (!res) send message saying that this sensor is not supported!
	}

	private void computeStatistics() {

		try {
			mX.setValidDataRange(mCurrNumOfSamples);
			mY.setValidDataRange(mCurrNumOfSamples);
			mZ.setValidDataRange(mCurrNumOfSamples);
		} catch (InvalidDataRangeException exception) {
			// this should never happen since we know mCurrNumOfSamples <= MAX_SAMPLES 
		}

		mX.calculateMean();
		mY.calculateMean();
		mZ.calculateMean();
	}

	private boolean isFall() {

		boolean ret = false;

		if      (mX.getVariance() > FALL_THRESHOLD) ret = true;
		else if (mY.getVariance() > FALL_THRESHOLD) ret = true;
		else if (mZ.getVariance() > FALL_THRESHOLD) ret = true;

		return ret;
	}

	private void stopDataCollection() {

		mSensorManager.unregisterListener(mSensorEventListener, 
		        mSensorManager.getDefaultSensor(android.hardware.Sensor.TYPE_ORIENTATION));

		mIsRunning = false;
	}

	private void prepareDataAndCallHandler() {

		SensorResponse response = new SensorResponse(
				mSensorInfo, new DefaultDouble3dData(mX.getVariance(), mY.getVariance(), mZ.getVariance(), mT));

		mLatestSensorResponse = response;
		mSensorListener.onNewResponse(response);
	}
}


