package fr.inria.arles.thinglib.devices.android;

import java.util.Map;

import android.content.Context;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.text.format.Time;

import fr.inria.arles.thinglib.datatypes.DefaultDouble3dData;
import fr.inria.arles.thinglib.devices.IntervalSensor;
import fr.inria.arles.thinglib.devices.ParameterErrorException;
import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.SensorListener;
import fr.inria.arles.thinglib.devices.SensorResponse;
import fr.inria.arles.thinglib.math.InvalidDataRangeException;
import fr.inria.arles.thinglib.math.SimpleStatistics;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;

/**
 * Since Android's Accelerometer, Magnetometer, and Orientation drivers operate
 * exactly the same way, this class contains code that is common to all of them.
 * Then it is left for each sensor that extends this class to also override the
 * missing methods. This means the getSensorInfo() and setSensingInterval()
 * methods.
 */
public abstract class InertialSensor implements IntervalSensor {

	// ******************************************************************************
	// Since this is one of the first sensors (or, as I'm calling it elsewhere,
	// "sensor driver") being written for Periscope, below you will find
	// comments that allows you to use this file as a template for writing your
	// own sensor drivers. 

	// ******************************************************************************
	// First, declare member variables

	SensorManager mSensorManager;
	SensorListener mSensorListener;
	Context context;
	boolean mIsRunning;
	//int mSensorNumber;
	private final int mSensorType;
	private final SensorInfo mSensorInfo;

	/**
	 * Specifies how old data can be to still be considered "fresh", in
	 * milliseconds.
	 */
	long mFreshnessMargin;

	double[] mXbuff, mYbuff, mZbuff;
	long mT;

	SimpleStatistics mX, mY, mZ;

	SensorResponse mLatestSensorResponse;

	int mMaxSamples, mCurrNumOfSamples;    
	int mSensingPeriod;


	// ******************************************************************************
	// Declare the constructor

	/** 
	 * Generic constructor for interial sensors. Differently from most other
	 * Sensor constructors, this one takes an additional parameter "sensorType".
	 * This parameter is a side-effect of the way the intertial sensors are
	 * implemented, and does not reflect how sensor drivers should be implemented in
	 * general. Normally, sensor drivers need only implement the
	 * single-parameter constructor.
	 *
	 * @param scope The Android Context (wrapped in an OuterScopeWrapper) for the
	 * activity or service that is calling this constructor. The reason why this
	 * is included here is because some of the sensors supported by Android must
	 * be controlled through the context.
	 *
	 * @param sensorType The type of intertial sensor being initialized (must be
	 * one of android.hardware.Sensor.TYPE_ACCELEROMETER, TYPE_MAGNETOMETER, or
	 * TYPE_ORIENTATION).
	 */
	public InertialSensor(OuterScopeWrapper scope, SensorInfo sensorInfo, int sensorType) {

		context = (Context) scope.getContents();

		mIsRunning = false;
		mSensorInfo = sensorInfo;
		mSensorType = sensorType;
		mFreshnessMargin = 3000; // in milliseconds
		mSensorManager = (SensorManager) context.getSystemService(Context.SENSOR_SERVICE);

		mX = new SimpleStatistics();
		mY = new SimpleStatistics();
		mZ = new SimpleStatistics();
		mT = -1;

		// set sensing interval to defaults
		// (must come after initialization of mX, mY, etc.)
		setSensingInterval(-1, -1);
	}


	// ******************************************************************************
	// Implement all the methods required by the superclass

	@Override
	public SensorResponse getData() {

		return mLatestSensorResponse;
	}

	@Override
	public void setSensingInterval(long sensingPeriod, int totalNumberOfSamples) {

		mMaxSamples = (totalNumberOfSamples > 0)? totalNumberOfSamples : InertialSensorInfo.preferredNumberOfSamples;

		// [Thiago] Since the sensing period has to agree with the for valued
		// allowed by Android, and I haven't implemented the validSensingPeriods
		// property of SensorInfo yet, I'm just going to set mSensingPeriod to
		// some value myself.
		mSensingPeriod = SensorManager.SENSOR_DELAY_GAME;

		afterSetSensingInterval();
	}

	@Override
	public void getData(SensorListener handler) {

		mSensorListener = handler;
		triggerDataCollection();
	}

	@Override
	public boolean hasFreshData() {

		Time time = new Time();
		time.setToNow();

		// TODO: Make sure we're handling timezones correctly (toMillis argument)
		if (mLatestSensorResponse.payload.timestamp > time.toMillis(false) - mFreshnessMargin) return true;

		return false;
	}

	@Override
	public void cancelCurrentTask() {

		if (mIsRunning) stopDataCollection();
	}

	@Override
	public boolean isSensing() {

		return mIsRunning;
	}
	

	@Override
	public Object getParameter(String key) 
	throws ParameterErrorException {

		// This method does nothing, since this class has no parameters
		throw new ParameterErrorException("This class has no parameters!");
	}

	@Override
	public void setParameter(String key, Object value)
	throws ParameterErrorException {
		
		// This method does nothing, since this class has no parameters
		throw new ParameterErrorException("This class has no parameters!");
	}

	@Override
	public Map<String, Object> getParameters() {

		// This method does nothing, since this class has no parameters
		return null;
	}	
		
	// NOTE: The few methods from Sensor that we do not override here are
	// actually being overridden in the classes that extend InertialSensor
	// (which is itself abstract, explaining why this is OK). For instance, look
	// inside AccelerometerSensor.java to see what I'm talking about.


	// ******************************************************************************
	// Implementation details

	private final SensorEventListener mSensorEventListener = new SensorEventListener() {
		/**
		 * This listener is called whenever there's a new inertial reading available.
		 */
		@Override
		public void onSensorChanged(SensorEvent event) {

			if (event.sensor.getType() != mSensorType) return;

			if (mCurrNumOfSamples < mMaxSamples) {

				if (mCurrNumOfSamples == 0) mT = event.timestamp;

				mXbuff[mCurrNumOfSamples] = event.values[0];
				mYbuff[mCurrNumOfSamples] = event.values[1];
				mZbuff[mCurrNumOfSamples] = event.values[2];

				mCurrNumOfSamples++;

				if (mCurrNumOfSamples == mMaxSamples) {
					stopDataCollection();
					computeStatistics();
					prepareDataAndCallHandler();
				}
			}
		}

		@Override
		public void onAccuracyChanged(android.hardware.Sensor sensor, int accuracy) {}
	};    

	protected void afterSetSensingInterval() {
		mXbuff = new double[mMaxSamples];
		mYbuff = new double[mMaxSamples];
		mZbuff = new double[mMaxSamples];
		mX.setData(mXbuff);
		mY.setData(mYbuff);
		mZ.setData(mZbuff);
	}

	private void triggerDataCollection() {

		mIsRunning = true;
		mCurrNumOfSamples = 0;
		
		boolean res = mSensorManager.registerListener(mSensorEventListener, 
		        mSensorManager.getDefaultSensor(mSensorType),
		        mSensingPeriod);

		// TODO IMPORTANT:
		// if (!res) send message saying that this sensor is not supported!
	}

	private void computeStatistics() {
		try {
			mX.setValidDataRange(mCurrNumOfSamples);
			mY.setValidDataRange(mCurrNumOfSamples);
			mZ.setValidDataRange(mCurrNumOfSamples);
		} catch (InvalidDataRangeException exception) {
			// this should never happen since we know mCurrNumOfSamples <= mMaxSamples
		}

		mX.calculateMean();
		mY.calculateMean();
		mZ.calculateMean();
	}

	private void stopDataCollection() {

		mSensorManager.unregisterListener(mSensorEventListener, 
		        mSensorManager.getDefaultSensor(mSensorType));

		mIsRunning = false;
	}

	private void prepareDataAndCallHandler() {

		SensorResponse response = new SensorResponse(
				mSensorInfo, new DefaultDouble3dData(mX.getMean(), mY.getMean(), mZ.getMean(), mT));

		mLatestSensorResponse = response;
		mSensorListener.onNewResponse(response);
	}

	@Override
	public SensorInfo getSensorInfo() {
		return mSensorInfo;
	}
}

