package fr.inria.arles.thinglib.devices.android;

import java.util.Map;

import android.content.Context;
import android.text.format.Time;
import android.util.Log;

import fr.inria.arles.thinglib.datatypes.DefaultDoubleNoiseLevelData;
import fr.inria.arles.thinglib.devices.IntervalSensor;
import fr.inria.arles.thinglib.devices.ParameterErrorException;
import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.SensorListener;
import fr.inria.arles.thinglib.devices.SensorResponse;
import fr.inria.arles.thinglib.devices.android.AudioReader.Listener;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;

public class SoundSensor implements IntervalSensor {	
	
	private static final int NOT_VALID_FUNCTION = -1;
	long mFreshnessMargin = 10000;
	SensorResponse message;
	SensorListener mhandler;
	AudioReader mAudioReader;
	double mCurrentSound;
	long mCurrentTimeStamp;
	int readError = AudioReader.Listener.ERR_OK;
	int mSampleRate;
	int mNumSamples;
	SensorInfo mSensorInfo;

	private static final String TAG = "SoundSensor";
		
	public SoundSensor(OuterScopeWrapper scope, SensorInfo sensorInfo) {
		Context context = (Context) scope.getContents();
		mAudioReader = new AudioReader();
		mSensorInfo = sensorInfo;
	}
	
	@Override
	public SensorInfo getSensorInfo() {
		return mSensorInfo;
	}

	@Override
	/**
	 * In the AudioRecorder documentation 
	 * (http://developer.android.com/reference/android/media/AudioRecord.html)
	 * it is stated that "44100Hz is currently the only rate that is guaranteed to work on all devices".
	 * 
	 * So, for the SoundSensor, the sampling rate (totalNumberOfSamples/periodInMicroSeconds) 
	 * will always be 44100
	 */
	public void setSensingInterval(long periodInMicroSeconds, int totalNumberOfSamples) {
		// mSampleRate = (int)(1000000/periodInMicroSeconds); // Converted into Frequncy
		/// mNumSamples = totalNumberOfSamples;
	}		

	// For sound measuring , this interface should not be implemented.
	@Override
	public SensorResponse getData() {
		Log.e(TAG, "getData() is not a valid function for SoundSensor Driver");
		SensorResponse message = new SensorResponse(mSensorInfo, NOT_VALID_FUNCTION ,new DefaultDoubleNoiseLevelData(0,0));
		return message;
	}

	/* User call this interface with Handler.
	 * Copy  handler.
	 * Call  SoundCollection Method.
	 */	
	@Override
	public void getData(SensorListener handler) {
		// Copy Handler.
		mhandler = handler;
		// Call SoundData Collection Method.
		
		Listener mAudioListener  = new AudioReader.Listener() {
			
			@Override
			public void onReadError(int error) {
				Log.e(TAG,"AudioReader.Listerer onReadError()");
				//On Read Complete-- stop Reading data. -- no need to do this, since the reader 
				//stops itself before calling this method.
				//mAudioReader.stopReader();
				// Read the error
				readError = error;
				// do packing in SensorResponse
				//TODO change back to null. Need to handle nulls better in .copy()
				message = new SensorResponse(mSensorInfo, readError, 
						new DefaultDoubleNoiseLevelData(-1, -1));
				//Send it to  user via handler.
				mhandler.onNewResponse(message);			
			}		
			

			@Override
			public void onReadComplete(short[] buffer) {
				Log.e(TAG,"AudioReader.Listerer onReadComplete()");
				// No need to stop the reader. The reader stops itself before calling us.
				// mAudioReader.stopReader();
				
					
				// Calculate the Signal power of "buffer" data.
				//Log.d("SOUND_SENSOR", "median value of buffer: "+c/(short)buffer.length);
				mCurrentSound = processAudio(buffer);
				short c = 0;
				for(int i = 0; i < buffer.length; i++)
					c += buffer[i];
				
				mCurrentTimeStamp = System.currentTimeMillis();			
				// Do Packing and send data to Handler 
				
				message = new SensorResponse(mSensorInfo, 
						new DefaultDoubleNoiseLevelData(mCurrentSound, mCurrentTimeStamp));
				mhandler.onNewResponse(message);
				}
			//}		
		};	
		
		mAudioReader.startReader(mSampleRate, mNumSamples, mAudioListener);
	}	

	@Override
	public boolean hasFreshData() {
		
		Time time = new Time();
		time.setToNow();		
		if(mCurrentTimeStamp > time.toMillis(false) - mFreshnessMargin ) 
			return true;	
		
		return false;
	}


	@Override
	public void cancelCurrentTask() {
		Log.e(TAG, "cancelCurrentTask()");
		stopDataCollection();		
	}	

	@Override
	public boolean isSensing() {
		//  TODO: implement this
		return true;
	}

	@Override
	public Object getParameter(String key) 
	throws ParameterErrorException {

		// This method does nothing, since this class has no parameters
		throw new ParameterErrorException("This class has no parameters!");
	}

	@Override
	public void setParameter(String key, Object value)
	throws ParameterErrorException {
		
		// This method does nothing, since this class has no parameters
		throw new ParameterErrorException("This class has no parameters!");
	}

	@Override
	public Map<String, Object> getParameters() {

		// This method does nothing, since this class has no parameters
		return null;
	}
	
	private void stopDataCollection(){
		mAudioReader.stopReader();		
	}
	

	/**
     * Audio Processing. 
     * @param   buffer      Audio data that was just read.
     */
    private final double processAudio(short[] buffer) {
    	double currentPower;
        // Process the buffer.  While reading it, it needs to be locked.
        synchronized (buffer) {
            // Calculate the power now, while we have the input buffer; 
        	final int len = buffer.length;            
            
           // calculate the signal power.
           currentPower = calculatePowerDb(buffer, 0, len);  
            
        }           	
        	return currentPower;        
    }  
    

    /**
     * Calculate the power of the given input signal.
     * 
     * @param   sdata       Buffer containing the input samples to process.
     * @param   off         Offset in sdata of the data of interest.
     * @param   samples     Number of data samples to process.
     * @return              The calculated power in dB relative to the maximum
     *                      input level; hence 0dB represents maximum power,
     *                      and minimum power is about -95dB.  Particular
     *                      cases of interest:
     *                      <ul>
     *                      <li>A non-clipping full-range sine wave input is
     *                          about -2.41dB.
     *                      <li>Saturated input (heavily clipped) approaches
     *                          0dB.
     *                      <li>A low-frequency fully saturated input can
     *                          get above 0dB, but this would be pretty
     *                          artificial.
     *                      <li>A really tiny signal, which only occasionally
     *                          deviates from zero, can get below -100dB.
     *                      <li>A completely zero input will produce an
     *                          output of -Infinity.
     *                      </ul>
     *                      <b>You must be prepared to handle this infinite
     *                      result and results greater than zero,</b> although
     *                      clipping them off would be quite acceptable in
     *                      most cases.
     */
    public final static double calculatePowerDb(short[] sdata, int off, int samples) {
        // Calculate the sum of the values, and the sum of the squared values.
        // We need longs to avoid running out of bits. 
    	
    	// This fudge factor is added to the output to make a realistically
        // fully-saturated signal come to 0dB.  Without it, the signal would
        // have to be solid samples of -32768 to read zero, which is not
        // realistic.  This really is a fudge, because the best value depends
        // on the input frequency and sampling rate.  We optimise here for
        // a 1kHz signal at 16,000 samples/sec.
        float FUDGE = 0.6f;
    	   	
    	// Maximum signal amplitude for 16-bit data.
        float MAX_16_BIT = 32768;
        double sum = 0;
        double sqsum = 0;
        for (int i = 0; i < samples; i++) {
            final long v = sdata[off + i];
            sum += v;
            sqsum += v * v;
        }        
       
        double power = (sqsum - sum * sum / samples) / samples;

        // Scale to the range 0 - 1.
        power /= MAX_16_BIT * MAX_16_BIT;

        // Convert to dB, with 0 being max power.  Add a fudge factor to make
        // a "real" fully saturated input come to 0 dB.
        return Math.log10(power) * 10f + FUDGE;
    }
}

