package fr.inria.arles.thinglib.math;

public class SimpleStatistics {

	private double[] mData;
	private int mDataRange;
	private boolean mHaveData;

	private double mMax, mMin, mMean, mVar;
	private boolean mHaveMax, mHaveMin, mHaveMean, mHaveVar;
	

	/**
	 * Constructor that creates a new SimpleStatistics object where all fields
	 * are initialized to Double.NaN.
	 */
	public SimpleStatistics() {
		mHaveData = false;
		resetStatistics();
	}


	/**
	 * Constructor that creates a new SimpleStatistics object where all fields
	 * are filled according to the max, min, mean, and variance of a given
	 * dataset.
	 *
	 * @param data An array of values for which statistics will be calculated.
	 */
	public SimpleStatistics(final double[] data) {
		setData(data);
	}


	/**
	 * Constructor that creates a new SimpleStatistics object where all fields
	 * are filled according to the max, min, mean, and variance of a given
	 * dataset.
	 *
	 * @param data An array of values for which statistics will be calculated.
	 *
	 * @param dataRange The number of items that should be considered when
	 * calculating statistics, starting from item 0. 
	 * 
	 * @throws InvalidDataRangeException 
	 */
	public SimpleStatistics(final double[] data, int dataRange) throws InvalidDataRangeException {
		setData(data, dataRange);
	}


	/**
	 * Sets the dataset whose statistics will be computed. 
	 *
	 * @param data An array of values for which statistics will be calculated.
	 */
	public void setData(final double[] data) {
		try {
			setData(data, data.length);
		}
		catch (InvalidDataRangeException exception) {
			// should never happen			
		}
	}


	/**
	 * Sets the dataset whose statistics will be computed. 
	 *
	 * @param data An array of values for which statistics will be calculated.
	 *
	 * @param dataRange The number of items that should be considered when
	 * calculating statistics, starting from item 0. 
	 * 
	 * @throws InvalidDataRangeException 
	 */
	public void setData(final double[] data, int dataRange) throws InvalidDataRangeException {

		mData = data;
		mHaveData = true;

		if (mDataRange < 0 || mData.length < mDataRange)
			throw new InvalidDataRangeException();

		mDataRange = dataRange;
		resetStatistics();
	}


	/**
	 * Sets the number of items that should be considered when calculating
	 * statistics, starting from item 0. This method must be called *after*
	 * setData() or a data-setting constructor is called.
	 * 
	 * @throws InvalidDataRangeException 
	 */
	public void setValidDataRange(int dataRange) throws InvalidDataRangeException {
		//if (!mHaveData) throw new DatasetEmptyException();
		if (dataRange < 0 || mData.length < dataRange) throw new InvalidDataRangeException();
		mDataRange = dataRange;
		resetStatistics();
	}


	/**
	 * Sets all fields to Double.NaN.
	 */
	public void resetStatistics() {
		mMax  = Double.NaN;
		mMin  = Double.NaN;
		mMean = Double.NaN;
		mVar  = Double.NaN;
		mHaveMax  = false;
		mHaveMin  = false;
		mHaveMean = false;
		mHaveVar  = false;
	}


	/**
	 * Simple method for constructing a new object and calculting its stats in
	 * one go.
	 *
	 * @param data An array of values for which statistics will be calculated.
	 *
	 * @param dataRange The number of items that should be considered when
	 * calculating statistics, starting from item 0. 
	 * 
	 * @throws InvalidDataRangeException 
	 */
	public static SimpleStatistics produce(final double[] data, int dataRange) throws InvalidDataRangeException {

		SimpleStatistics s = new SimpleStatistics(data, dataRange);
		s.calculateAllStats();
		return s;
	}

	/**
	 * Simple method for constructing a new object and calculting its stats in
	 * one go.
	 *
	 * @param data An array of values for which statistics will be calculated.
	 */
	public static SimpleStatistics produce(final double[] data) {
		try {
			return produce(data, data.length);
		}
		catch (InvalidDataRangeException exception) {
			// should never happen
			return new SimpleStatistics();
		}	
	}


	/**
	 * Throws exceptions if this instance does not have enough information to
	 * compute statistics or if the selected data range is out of bounds.
	 * @throws DatasetEmptyException 
	 */
	protected void checkHaveData() throws DatasetEmptyException {
		if (!mHaveData) {
			throw new DatasetEmptyException();
		}
	}	


	/**
	 * Returns the maximum value found in the dataset, calculating it if
	 * necessary.
	 */
	public double getMax() {
		if (!mHaveMax) calculateMax();
		return mMax;
	}


	/**
	 * Returns the minimum value found in the dataset, calculating it if
	 * necessary.
	 */
	public double getMin() {
		if (!mHaveMin) calculateMin();
		return mMin;
	}


	/**
	 * Returns the mean of all values in the dataset, calculating it if
	 * necessary.
	 */
	public double getMean() {
		if (!mHaveMean) calculateMean();
		return mMean;
	}


	/**
	 * Returns the variance of all values in the dataset, calculating it if
	 * necessary.
	 */
	public double getVariance() {
		if (!mHaveVar) calculateVariance();
		return mVar;
	}


	/**
	 * Calculates the maximum value found in the dataset but just saves it
	 * internally rather than returning it. To fetch the value after
	 * calculating it, use getMax().
	 */
	public void calculateMax() {
		calculateMaxMinMean();
	}


	/**
	 * Calculates the min value found in the dataset but just saves it
	 * internally rather than returning it. To fetch the value after
	 * calculating it, use getMin().
	 */
	public void calculateMin() {
		calculateMaxMinMean();
	}


	/**
	 * Calculates the mean value found in the dataset but just saves it
	 * internally rather than returning it. To fetch the value after
	 * calculating it, use getMean().
	 */
	public void calculateMean() {
		calculateMaxMinMean();
	}


	protected void calculateMaxMinMean() {

		if (!mHaveData) return;

		double x = mData[0];
        double max = x;
        double min = x;
        double sum = x;
		double mean;

		// calculate mean, max, min

        for (int i = 1; i < mDataRange; i++) {
			x = mData[i];
            if (x > max) max = x;
            if (x < min) min = x;
            sum += x;
        }

        mean = ((double) sum) / mDataRange;
		
		mMax  = max;
		mMin  = min;
		mMean = mean;
		mHaveMax  = true;
		mHaveMin  = true;
		mHaveMean = true;
	}


	/**
	 * Calculates the variance of the values in the dataset but just saves it
	 * internally rather than returning it. To fetch the value after
	 * calculating it, use getVariance().
	 */
	public void calculateVariance() {

		if (!mHaveData) return;

		double e;
		double eSum = 0;  // sum of the "errors" --> Sum( (x - mean) )
		double eeSum = 0; // sum of the "errors" squared --> Sum( (x - mean)^2 )

        for (int i = 0; i < mDataRange; i++) {
			e = mData[i] - mMean;
			eSum  += e;
			eeSum += e*e;
        }

		mVar = (eeSum - eSum*eSum/mDataRange) / (mDataRange - 1);
		mHaveVar = true;
	}



	/**
	 * Calculates all supported statistics about the dataset but saves it
	 * internally rather than returning it.
	 */
	protected void calculateAllStats() {
		calculateMaxMinMean();
		calculateVariance();
	}
}
