package fr.inria.arles.thinglib.platforms.android;

import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.content.pm.PackageManager;
import android.content.pm.ResolveInfo;
import android.os.IBinder;
import android.os.RemoteException;
import android.util.Log;

import java.io.IOException;
import java.io.StreamCorruptedException;
import java.util.HashMap;
import java.util.List;
import java.util.ServiceConfigurationError;
import java.util.Vector;

import fr.inria.arles.thinglib.core.DuplicateDeviceException;
import fr.inria.arles.thinglib.core.Messaging;
import fr.inria.arles.thinglib.devices.SensorInfo;
import fr.inria.arles.thinglib.devices.android.SensorAidlInterface;
import fr.inria.arles.thinglib.platforms.OuterScopeWrapper;

public class Daemon extends fr.inria.arles.thinglib.core.Daemon {

	public static final String PLUGIN_INTENT_STRING = "fr.inria.arles.thinglib.intents.GET_THING";
	protected final OuterScopeWrapper mOuterScope;
	protected int mSensorsLeftToProcess;
	protected DaemonService mDaemonService;
	

    /** Keeps track of all current registered clients. */
    protected HashMap<Integer, ClientAidlInterface> mClients;

	public Daemon(DaemonService daemonService) {
		super();
		mClients       = new HashMap<Integer, ClientAidlInterface>();
		mOuterScope    = new OuterScopeWrapper(daemonService);
		mDaemonService = daemonService;
	}

	private final DaemonAidlInterface.Stub mBinder = new DaemonAidlInterface.Stub() {

		@Override
		public void registerCallback(int clientAddress, ClientAidlInterface clientAidlObject) {
			
			mClients.put(clientAddress, clientAidlObject);
		}

		@Override
		public void handleMessageFromClient(int clientAddress, int messageType, byte[] messagePayload) {

			processMessage(clientAddress, messageType, messagePayload);
		}

		@Override
		public byte[] handleRpcFromClient(int clientAddress, int messageType, byte[] messagePayload) {

			return processMessage(clientAddress, messageType, messagePayload);
		}
	};

	public IBinder getBinder() {
		
		return mBinder;
	}

	@Override
	protected int registerClient() {

		// Override registration to make sure new clients get passed their OuterScope.

		int clientAddress = super.registerClient();
		setOuterScope(clientAddress, null);

		return clientAddress;
	}


	@Override
	public void setOuterScope(int clientHandle, byte[] messagePayload) {

		// ignore the passed outerScope, and instead use the one saved earlier
		mWorkers.get(clientHandle).setOuterScope(mOuterScope);
	}

	@Override
	protected void sendMessageToClient(int clientAddress, int messageType, byte[] messagePayload) {

		try {
			mClients.get(clientAddress).handleMessageFromDaemon(messageType, messagePayload);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void discoverSensors() {

		setDiscoveryState(STATE_DISCOVERY_IN_PROGRESS);

		Log.e("DAEMON", "Discovering sensors!");
		mSensorInfos = new Vector<SensorInfo>();

        PackageManager packageManager = mDaemonService.getPackageManager();
        Intent genericPluginIntent = new Intent(PLUGIN_INTENT_STRING);      
        List<ResolveInfo> resolveInfos = packageManager.queryIntentServices(genericPluginIntent, 0);
        
		mSensorsLeftToProcess = resolveInfos.size();        

		Intent pluginIntent;
		
		for (ResolveInfo resolveInfo : resolveInfos) {

			pluginIntent = new Intent(PLUGIN_INTENT_STRING);
			pluginIntent.addCategory(resolveInfo.serviceInfo.name);
			
			// mDaemonService.startService(pluginIntent);
			
			ServiceConnection mSensorConnection = new ServiceConnection() {

				@Override
				public void onServiceConnected(ComponentName className, IBinder binder) {

					SensorAidlInterface sensorService = SensorAidlInterface.Stub.asInterface(binder);

					try { 
						byte[] sensorInfoBytes = sensorService.getSensorInfo();
						SensorInfo sensorInfo = (SensorInfo) Messaging.toObject(sensorInfoBytes);
						installSensor(sensorInfo);
						
						mDaemonService.unbindService(this);
						
					} catch (ServiceConfigurationError e) {
						e.printStackTrace();
					} catch (DuplicateDeviceException e) {
						e.printStackTrace();
					} catch (RemoteException e) {
						e.printStackTrace();
					} catch (StreamCorruptedException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					} catch (ClassNotFoundException e) {
						e.printStackTrace();
					}
								
					mSensorsLeftToProcess--;
					
					if (mSensorsLeftToProcess <= 0) {
						
						if (mDiscoveryState != STATE_DISCOVERY_DONE) {
							setDiscoveryState(STATE_DISCOVERY_DONE);
						}
						
						// Alert the outside world that new sensors have been found
						triggerOnSensorsChanged();				
					}
				}

				@Override
				public void onServiceDisconnected(ComponentName name) {

					// Alert the outside world a sensors has been disconnected
					triggerOnSensorsChanged();
				}
			};
			
			mDaemonService.bindService(pluginIntent, mSensorConnection, Context.BIND_AUTO_CREATE);

		}

	}
}


