package fr.inria.arles.thinglib.semantics;

/** 
 * A SemanticClass is a class that can be represented and type-checked across
 * different-language implementations. Classes that extend SemanticClass must
 * define a unique ID (UID) and an interface, each represented by a String that
 * does not contain spaces (THIS IS IMPORTANT!).
 *
 * Although these strings can follow any convention, for public projects it is a
 * good idea to use reverse-domain-name notation (such as
 * "fr.inria.arles.MyUID"). In case of interface string, they can also consist
 * of real URLs that point to a schema definition (in a format that is yet to be
 * determined).
 */
public abstract class SemanticClass {

	/**
	 * Returns a string that uniquely identifies this class. Make sure this
	 * string contains version numbers, for future-proofing. 
	 *
	 * No spaces allowed.
	 */
	public abstract String getUid();

	/**
	 * Returns a string that uniquely identifies the natural interface followed
	 * by this class. Make sure this string contains version numbers, for
	 * future-proofing. 
	 *
	 * Note that the interface may be the same as the UID if this is the first
	 * class of its sort, in that it does not implement any existing interface.
	 *
	 * No spaces allowed.
	 */
	public abstract String getInterface();

	/**
	 * Read a property specified by this natural class, as defined by its
	 * interface.
	 */
	public Object getProperty(String propertyName) 
	throws NotImplementedException {
		throw new NotImplementedException(propertyName);
	}

	/**
	 * Set a property specified by this natural class, as defined by its
	 * interface.
	 */
	public void setProperty(String propertyName, Object value)
	throws NotImplementedException {
		throw new NotImplementedException(propertyName);
	}

	/* Convenience methods **********************************************/

	/**
	 * Returns true if this SensorInfo's UID string matches the given
	 * string.
	 *
	 * @param uid The UID to be tested.
	 */
	public boolean isIdentifiedBy(String uid) {
		return getUid().equals(uid);
	}

	/**
	 * Returns true if this SensorInfo's interface string matches the given
	 * string.
	 *
	 * @param interfaceStr A string representing the interface to be tested.
	 */
	public boolean implementsInterface(String interfaceStr) {
		return getInterface().equals(interfaceStr);
	}
}

