/******************************************************************
Author:  		Georgios Mathioudakis (georgios.mathioudakis@inria.fr)
Project: 		AndroidSensorAccessDemo
package: 		fr.inria.arles.thinglib.RESTAccess
Updated: 		
Status:  		Working Draft
Description: 	This class implements the WebSocket server used for the streaming feature. More information can be found in https://github.com/TooTallNate/Java-WebSocket
 ******************************************************************/

package fr.inria.arles.thinglib.websockets;

import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;
import org.java_websocket.WebSocket;
import org.java_websocket.handshake.ClientHandshake;
import org.java_websocket.server.WebSocketServer;
import org.json.JSONException;
import org.json.JSONObject;

import fr.inria.arles.thinglib.RESTAccess.MainActivity;
import fr.inria.arles.thinglib.RESTAccess.MediatorService;
import fr.inria.arles.thinglib.RESTAccess.StreamingResponseListener;
import android.os.Handler;
import android.util.Log;

public class WebSocket_Server extends WebSocketServer {

	Handler StopSensingHandler = new Handler();
	Runnable StopSensingSched = new Runnable() {

		@Override
		public void run() {
			/*
			 * If there are no subscribers stop the periodic sensing
			 */
			MediatorService.getController().stopSensing();
		}
	};

	private String TAG = "WebSocket_Server";
	private Map<WebSocket, String> _requests;

	public WebSocket_Server(int port) throws UnknownHostException {
		super(new InetSocketAddress(port));
		_requests = new HashMap<WebSocket, String>();
		Log.i(TAG, "WebSocket server on device is started on port: " + port);
	}

	public WebSocket_Server(InetSocketAddress address) {
		super(address);
	}

	@Override
	public void onOpen(WebSocket conn, ClientHandshake handshake) {
		Log.i(TAG, conn.getRemoteSocketAddress().getAddress().toString()
				+ " has been connected to the websocket server");
	}

	@Override
	public void onClose(WebSocket conn, int code, String reason, boolean remote) {
		Log.i(TAG, conn + " has been disconnected from the websocket server");
		/*
		 * When a client closes we should check if there are any other clients
		 * connected to the streaming server. If not it is safe to stop the
		 * periodic sensing. Because a client may be be restarted
		 * spontaneously(browser refresh) the close message arrives later than
		 * the new open message Thus, the onClose is called later than the
		 * onOpen causing the mediator to stop sensing despite the fact the
		 * there is an open connection To avoid this situation we schedule the
		 * checking for some seconds after the onClose event. This way if any
		 * other connection are opened meanwhile will be counted as well.
		 */
		StopSensingHandler.postDelayed(StopSensingSched, 2000);

		/*
		 * Unsubscribe the client from requests list
		 */
		_requests.remove(conn);
	}

	/*
	 * Use this method to get streaming requests
	 */
	@Override
	public void onMessage(WebSocket conn, String message) {
		System.out.println("Message received to WS server: " + message);
		/*
		 * Parse message to get the requests Message should be in the form
		 * {"sensorDataRequest":{"serviceName":"fr.inria.arles.sensor.location", "deviceID":"41a768f830e3119ea1e32f5532fc0e0fe2f270c8"}}
		 */
		String serviceRequested = "";
		int rate = 3000;
		try {
			JSONObject reqObj = new JSONObject(message);
			JSONObject request = reqObj.getJSONObject("sensorDataRequest");
			serviceRequested = request.getString("serviceName");
			// rate = Integer.valueOf(request.getString("rate"));
		} catch (JSONException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		if (!MediatorService.getController().IsSensing()) {
			MediatorService.getController().startSensing(rate);
			Log.i(TAG, "Mediator is sensing");
		}

		/*
		 * Store the connection and the service request to the _requests map in
		 * order to send the data when they are available
		 */
		_requests.put(conn, serviceRequested);

	}

	@Override
	public void onError(WebSocket conn, Exception ex) {
		ex.printStackTrace();
		if (conn != null) {
			Log.e(TAG, "WebSocket server ERROR");
		}
	}

	public void SendResponce(JSONObject response) {
		String responseService = "";
		try {
			JSONObject responseData = response
					.getJSONObject("sensorDataResponse");
			responseService = responseData.getString("serviceName");
		} catch (JSONException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		if (responseService.equals(""))
			return;

		if (_requests.containsValue(responseService)) {
			for (Map.Entry<WebSocket, String> pair : _requests.entrySet()) {
				if (pair.getValue().equals(responseService))
					pair.getKey().send(response.toString());
			}
		}

	}
}
