package org.ow2.choreos.registryclient.impl;

import android.util.Log;

import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;
import org.ow2.choreos.iots.common.Service;
import org.ow2.choreos.registryclient.interfaces.RegistryClient;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;


public class RestRegistryImpl implements RegistryClient {
    private static final String TAG = "RegistryImpl";
    String baseURL;

    public RestRegistryImpl(String urlString) {
        baseURL = urlString;
    }

    @Override
    /**
     * This method is called by the registration manager to find the number of services measuring concepts of interest
     * the number should be provided for each concept
     * @param myConcepts: the concepts of interest
     * @param locationOfInterest: the location at which concepts of interest should be measured
     * @param deviceId: the id of the device hosting the initial service. This is important
     * because it should be excluded when returning the number of services measuring each concept.
     * @return the HashMap containing the concept of interest and the number of registered services that
     * measure this concept.
     */
    public HashMap<String, Double> findNumberOfServicesReq(
            ArrayList<String> myConcepts, String locationOfInterest,
            String deviceId) {
        HashMap<String, Double> thisMap = new HashMap<String, Double>();
        String paramString = "?device_id=" + deviceId + "&global_region="
                + locationOfInterest;
        for (int i = 0; i < myConcepts.size(); i++) {
            paramString = paramString + "&concept=" + myConcepts.get(i);

        }

        URL url;
        HttpURLConnection conn = null;
        try {
            url = new URL(baseURL + "find_number_of_services" + paramString);
            conn = (HttpURLConnection) url.openConnection();

            BufferedReader in = null;
            in = new BufferedReader(new InputStreamReader(
                    conn.getInputStream()));
            String response = null;
            try {
                while ((response = in.readLine()) != null) {
                    System.out.println(response);
                    JSONParser myParser = new JSONParser();
                    try {
                        JSONObject responseObject = (JSONObject) myParser
                                .parse(response);
                        thisMap = (HashMap) responseObject.get("value");

                    } catch (ParseException e) {
                        System.out
                                .println("RestRegistryImpl.findNumberOfServices: returned object is incorrect");
                        e.printStackTrace();
                        return null;
                    }

                }
            } catch (IOException e) {
                System.out
                        .println("RestRegistryImpl.findNumberOfServices: could not return a correct response");
                e.printStackTrace();
                return null;
            }


        } catch (MalformedURLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        } catch (IOException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        } finally {
            conn.disconnect();
        }

        return thisMap;

    }

    @Override
/**
 * The method is called to perform the actual registration of the service in a device registry.
 * if the registry is remote, it invokes a web service.
 * @param ps: the service to register
 * @return true if the registration was successful and false otherwise.
 */
    public boolean registerServiceReq(Service ps) {
        boolean result = false;
        URL url;
        HttpURLConnection conn = null;
        try {
            url = new URL(baseURL + "register_service");
            conn = (HttpURLConnection) url.openConnection();
            conn.setRequestMethod("POST");
            conn.setDoInput(true);
            conn.setDoOutput(true);
        } catch (MalformedURLException e) {
            e.printStackTrace();
        } catch (ProtocolException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        BufferedWriter out;
        try {
            out = new BufferedWriter(new OutputStreamWriter(conn.getOutputStream()));

            // TODO send ontology namespace? (to be added to concept and sensor)
            out.write("address=" + ps.getServiceAddress() + "&");
            out.write("device_id=" + ps.getDeviceId() + "&");
            out.write("name=" + ps.getServiceName() + "&");
            out.write("type=" + ps.getServiceType() + "&");
            out.write("accuracy=" + ps.getServiceAccuracy() + "&");
            out.write("sensor=" + ps.getTypeOfServiceSensor() + "&");
            out.write("range=" + ps.getServiceRange() + "&");
            out.write("global_region=" + ps.getGlobalRegion().getLocationName() + "&");
            out.write("concept=" + ps.getConcept() + "&");
            out.write("dataType=" + ps.getMeasuredDataType() + "&");
            out.write("coordinateX=" + ps.getCoordinateX() + "&");
            out.write("coordinateY=" + ps.getCoordinateY() + "&");

			/* MISING ARGUMENTS */
            out.write("unit=" + ps.getUnit() + "&");
            out.write("estimated_path=" + ps.getEstimatedPathString() + "&");
            out.write("expiry_date=" + "100000");
            out.flush();
            out.close();
        } catch (IOException e) {
            Log.d(TAG, "RestRegistryImpl.registerServiceReq: could not write outputStream");
            e.printStackTrace();
            return false;
        }

        BufferedReader in = null;
        try {
            in = new BufferedReader(new InputStreamReader(conn.getInputStream()));

            String response;

            while ((response = in.readLine()) != null) {
                if (response.equals("true"))
                    result = true;
                else
                    result = false;
            }
            in.close();
        } catch (MalformedURLException e) {
            e.printStackTrace();
            return false;
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            conn.disconnect();
        }

        return result;
    }

    @Override
    /**
     * This method is called to remove the service from the registry if needed. If the registry is remote,
     * this method invokes a web service.
     * @param phoneService: the service to remove.
     */
    public void removeServiceReq(Service phoneService) {
        try {
            URL url = new URL(baseURL + "remove_service");
            HttpURLConnection conn = null;

            try {
                conn = (HttpURLConnection) url.openConnection();
                conn.setRequestMethod("POST");
                conn.setDoInput(true);
                conn.setDoOutput(true);
            } catch (IOException e) {
                System.out
                        .println("RestRegistryImpl.removeServiceReq: could not open connection");
                e.printStackTrace();
                return;
            }

            BufferedWriter out = null;
            try {
                out = new BufferedWriter(new OutputStreamWriter(
                        conn.getOutputStream()));
            } catch (IOException e) {
                System.out
                        .println("RestRegistryImpl.removeServiceReq: could not create outputStream");
                e.printStackTrace();
                return;
            }
            try {
                out.write("address=" + phoneService.getServiceAddress());

            } catch (IOException e) {
                System.out
                        .println("RestRegistryImpl.removeServiceReq: could not write ouputStream");
                e.printStackTrace();
                return;
            }
            try {
                out.flush();
            } catch (IOException e) {
                e.printStackTrace();
                return;
            }
            try {
                out.close();
            } catch (IOException e) {
                e.printStackTrace();
                return;
            }
            BufferedReader in = null;
            try {
                in = new BufferedReader(new InputStreamReader(
                        conn.getInputStream()));
            } catch (IOException e) {
                System.out
                        .println("RestRegistryImpl.removeServiceReq: could not get inputStream");
                e.printStackTrace();
                return;
            }

            String response;
            try {
                while ((response = in.readLine()) != null) {
                    System.out.println(response);
                }
            } catch (IOException e) {
                System.out
                        .println("RestRegistryImpl.removeServiceReq: could not read inputStream");
                e.printStackTrace();
                return;
            }
            try {
                in.close();
            } catch (IOException e) {
                e.printStackTrace();
                return;
            }

        } catch (MalformedURLException e) {
            System.out
                    .println("RestRegistryImpl.removeServiceReq: URL was malformed");
            e.printStackTrace();
            return;
        }

    }

}
